#Additional subroutines used inside cleanMaskUI.py
#Not really meant to be stand alone utilities.

import numpy as np
import scipy
import tkinter as tk
import tkinter.ttk as ttk
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
#Try to import the default matplotlib toolbar
try: #for matplotlib 3.x
    from matplotlib.backends.backend_tkagg import NavigationToolbar2Tk
    _mplToolbar=3
except ImportError:  #If that fails this probably isn't matplotlib 3.x
    try: #for matplotlib 2.x
        from matplotlib.backends.backend_tkagg import NavigationToolbar2TkAgg
        _mplToolbar=2
    except ImportError: #if that fails, just don't use this extra toolbar
        _mplToolbar=0

from . import profileLSD
from . import mask as maskTools
from . import obsSpec

def makeWin(fig, ax, mask, obs, lsdp, lsdProf, pltMaskU, pltMaskN,
            pltMaskF, pltModelI, excludeRanges, outExcludeName, fitDepthFlags):
    #Build GUI with tkinter
    root = tk.Tk(className='Clean Masks')
    #the className seems to set an icon title (at least in Ubuntu)
    root.title("Clean Masks")

    ##For an icon image (used by window managers)
    #try:
    #    imgIcon = tk.Image("photo", file="./iconNorm.png")
    #    root.iconphoto(True, imgIcon)
    #except:
    #    pass

    #Generate a tk.DrawingArea using matplotlib's Tk backend,
    #from the given figure
    canvas = FigureCanvasTkAgg(fig, master=root)
    canvas.draw() #this draw call may not be strictly necessary? 
    canvasWidget = canvas.get_tk_widget()
    canvasWidget.pack(side=tk.TOP, fill=tk.BOTH, expand=1)
    canvas.figure.tight_layout()  #try to use matplotlib's tighter boarders

    #Add a matplotlib's standard toolbar.
    if _mplToolbar > 0:
        if _mplToolbar == 2:
            toolbarMpl = NavigationToolbar2TkAgg(canvas, root)
        if _mplToolbar == 3:
            toolbarMpl = NavigationToolbar2Tk(canvas, root)
        #Note, making the toolbar calls the parent container's .pack()
        toolbarMpl.update() #not strictly necessary
        canvas.get_tk_widget().pack(side=tk.TOP, fill=tk.BOTH, expand=1)

    #Add custom UI controls for the plot
    
    #Create a little event manager object for processing
    #callbacks for view changing buttons
    viewM = viewFuncs(ax, canvas)
    #Make a frame to hold our tool buttons,
    #and allow us to use use the tk grid geometry manager for them
    tools = ttk.Frame(root, padding="3 3 3 3")
    tools.pack(side=tk.BOTTOM, fill=tk.X)
    #and put view changing buttons in their own frame
    viewtools = ttk.Frame(tools)
    viewtools.grid(row=3, column=0, columnspan=8, sticky=tk.W, padx=2)
    #
    butAuto = ttk.Button(master=viewtools, text="auto",
                         command=viewM.autoScale, width=7)
    ToolTip(butAuto, "Auto-scale axes ranges [a]")
    butAuto.grid(row=0, column=0, sticky=(tk.E,tk.W))
    butAutoY = ttk.Button(master=viewtools, text="auto-y",
                          command=viewM.autoScaleY, width=7)
    ToolTip(butAutoY, "Auto-scale y-axis range [A]")
    butAutoY.grid(row=0, column=1, sticky=(tk.E,tk.W))
    butZin = ttk.Button(master=viewtools, text="zoom",
                        command=viewM.zoomRec, width=5)
    ToolTip(butZin, "Zoom to selected region [z]")
    butZin.grid(row=0, column=2, sticky=(tk.E,tk.W))
    butZin = ttk.Button(master=viewtools, text="Zin",
                        command=viewM.zoomIn, width=5)
    ToolTip(butZin, "Zoom in [i]")
    butZin.grid(row=0, column=3, sticky=(tk.E,tk.W))
    butZout = ttk.Button(master=viewtools, text="Zout",
                         command=viewM.zoomOut, width=5)
    ToolTip(butZout, "Zoom out [o]")
    butZout.grid(row=0, column=4, sticky=(tk.E,tk.W))
    butLeft = ttk.Button(master=viewtools, text="\u2190",
                         command=viewM.panLeft, width=3)
    ToolTip(butLeft, "Pan left [arrow key]")
    butLeft.grid(row=0, column=5, sticky=(tk.E,tk.W))
    butRight = ttk.Button(master=viewtools, text="\u2192",
                          command=viewM.panRight, width=3)
    ToolTip(butRight, "Pan right [arrow key]")
    butRight.grid(row=0, column=6, sticky=(tk.E,tk.W))
    butUp = ttk.Button(master=viewtools, text="\u2191",
                       command=viewM.panUp, width=3)
    ToolTip(butUp, "Pan up [arrow key]")
    butUp.grid(row=0, column=7, sticky=(tk.E,tk.W))
    butDown = ttk.Button(master=viewtools, text="\u2193",
                         command=viewM.panDown, width=3)
    ToolTip(butDown, "Pan down [arrow key]")
    butDown.grid(row=0, column=8, sticky=(tk.E,tk.W))
    #Allow .grid() managed buttons to expand/contract if necessary
    viewtools.rowconfigure(0, weight=1)
    viewtools.columnconfigure(0, weight=1)
    viewtools.columnconfigure(1, weight=1)
    viewtools.columnconfigure(2, weight=1)
    viewtools.columnconfigure(3, weight=1)
    viewtools.columnconfigure(4, weight=1)
    viewtools.columnconfigure(5, weight=1)
    viewtools.columnconfigure(6, weight=1)
    viewtools.columnconfigure(7, weight=1)
    viewtools.columnconfigure(8, weight=1)
    
    #Set some general key bindings
    root.bind('<Key-Left>', viewM.panLeft)
    root.bind('<Key-Right>', viewM.panRight)
    root.bind('<Key-Up>', viewM.panUp)
    root.bind('<Key-Down>', viewM.panDown)
    root.bind('a', viewM.autoScale)
    root.bind('A', viewM.autoScaleY)
    root.bind('z', viewM.zoomRec)
    root.bind('i', viewM.zoomIn)
    root.bind('o', viewM.zoomOut)


    #let the .grid() managed cells expand
    tools.columnconfigure(6, weight=10)
    
    #Set regions to be used for adjusting line depths
    selectFitDepth = uiSelectFitDepth(canvas, mask, pltMaskU, pltMaskN, pltMaskF,
                                      excludeRanges, fitDepthFlags)
    butSelRange = ttk.Button(master=tools, text='select lines\nto fit depth',
                             command=selectFitDepth.runSpanSelect)
    ToolTip(butSelRange, 'Select lines to fit for depth in the mask')
    butSelRange.grid(row=0, column=0, sticky=tk.W, padx=2, pady=2)
    #Set regions to not be used for adjusting line depths
    unselectFitDepth = uiUnselectFitDepth(canvas, mask, pltMaskU, pltMaskN, pltMaskF,
                                          excludeRanges, fitDepthFlags)
    butUnselRange = ttk.Button(master=tools, text='unselect lines\nto fit depth',
                             command=unselectFitDepth.runSpanSelect)
    ToolTip(butUnselRange, 'Unselect lines to fit for depth in the mask')
    butUnselRange.grid(row=0, column=1, sticky=tk.W, padx=2, pady=2)
    #Link the select and unselect buttons so they can turn each other off,
    #done after creating the include & exclude buttons so they all can be linked
    #Apply the depth fitting
    fitDepthsM = fitDepths(mask, obs, lsdp, lsdProf, fitDepthFlags, root,
                           canvas, pltMaskU, pltMaskN, pltMaskF)
    butFitDepths = ttk.Button(master=tools, text='fit\ndepths',
                               command=fitDepthsM.runFit)
    ToolTip(butFitDepths, 'Fit the selected line depths, using the current LSD profile and observation')
    butFitDepths.grid(row=0, column=3, sticky=tk.W, padx=2, pady=2)
    #Undo the depth fitting
    undoDepthsM = undoDepths(mask, fitDepthFlags, canvas,
                             pltMaskU, pltMaskN, pltMaskF)
    butUndoDepths = ttk.Button(master=tools, text='undo\nfit',
                               command=undoDepthsM.undo)
    ToolTip(butUndoDepths, 'Undo all line depth fitting of currently selected lines, reverting them to the initial mask depths')
    butUndoDepths.grid(row=0, column=4, sticky=tk.W, padx=2, pady=2)
    
    speratorFitting = ttk.Separator(tools, orient=tk.VERTICAL)
    speratorFitting.grid(row=0, column=5, sticky='NS', padx=2, pady=2)

    #Set regions to be included in the mask
    incRangeM = uiIncludeRange(canvas, mask, pltMaskU, pltMaskN, pltMaskF,
                               excludeRanges, fitDepthFlags)
    butIncRange = ttk.Button(master=tools, text='include\nlines',
                             command=incRangeM.runSpanSelect)
    ToolTip(butIncRange, 'Include selected lines in the mask')
    butIncRange.grid(row=0, column=7, sticky=tk.E, padx=2, pady=2)
    #Set regions to be excluded in the mask
    excRangeM = uiExcludeRange(canvas, mask, pltMaskU, pltMaskN, pltMaskF,
                               excludeRanges, fitDepthFlags)
    butExcRange = ttk.Button(master=tools, text='exclude\nlines',
                             command=excRangeM.runSpanSelect)
    ToolTip(butExcRange, 'Exclude selected lines from the mask')
    butExcRange.grid(row=0, column=8, sticky=tk.E, padx=2, pady=2)
    #Link the include and exclude buttons so they can turn each other off,
    #done in the next block where all related buttons are linked
    #Save the exclude ranges to a text file
    saveRangesM = saveRanges(outExcludeName, excludeRanges)
    butSaveRanges = ttk.Button(master=tools, text='save\nranges',
                               command=saveRangesM.saveToFile)
    ToolTip(butSaveRanges, 'Save the excluded wavelength ranges to the {:} file'.format(outExcludeName))
    butSaveRanges.grid(row=0, column=9, sticky=tk.E, padx=2, pady=2)

    #Link the select, unselect, include, and exclude buttons
    #so they can all turn each other off, and have only one mode active
    selectFitDepth.linkButton(butSelRange, [unselectFitDepth, incRangeM, excRangeM])
    unselectFitDepth.linkButton(butUnselRange, [selectFitDepth, incRangeM, excRangeM])
    incRangeM.linkButton(butIncRange, [excRangeM, selectFitDepth, unselectFitDepth])
    excRangeM.linkButton(butExcRange, [incRangeM, selectFitDepth, unselectFitDepth])
    
    #Modify LSD control parameters
    modifyLSDparM = modifyLSDpar(root, lsdp)
    butModLSD = ttk.Button(master=tools, text='LSD param.',
                              command=modifyLSDparM.openWindow)
    ToolTip(butModLSD, 'Modify the parameters related to the LSD calculation')
    butModLSD.grid(row=3, column=8, sticky=tk.E, padx=2)
    
    #Update the LSD calculation and the plotted spectrum
    updateLSDM = updateLSD(canvas, root, mask, lsdp, lsdProf, pltModelI)
    butUpdateLSD = ttk.Button(master=tools, text='update LSD',
                              command=updateLSDM.rerunLSD)
    ToolTip(butUpdateLSD, 'Save the mask, run LSD, and update the model spectrum')
    butUpdateLSD.grid(row=3, column=9, sticky=(tk.N, tk.S, tk.E, tk.W), padx=2)
    
    #Run the main loop!
    root.mainloop()


def combineRanges(ranges1, ranges2):
    #ranges1 and ranges2 are lists of ranges, each containing 2 entry lists 
    #of start and end wavelengths for a range.
    #This combines the two sets of ranges and removes overlap by merging ranges.
    #combine ranges and sort by starting wavelength
    rangesT = sorted(ranges1 + ranges2, key=lambda ran: ran[0])
    #recursively merge adjacent entries
    update = True
    while update == True:
        update = False
        skip = False
        rangesO = []
        #Loop over all but the last entry (compare this entry with the next)
        for i, r1 in enumerate(rangesT[:-1]):
            #Skip this entry if it's been merged with the previous
            if skip:
                skip = False
                continue
            r2 = rangesT[i+1]
            #if there is overlap merge these windows
            if r1[1] >= r2[0]:
                update = True
                #if the next range ends after this range
                if r2[1] > r1[1]:
                    rangesO += [[r1[0], r2[1]]]
                    skip = True
                #if the next range ends before this range (is smaller)
                else:
                    rangesO += [[r1[0], r1[1]]]
                    skip = True
            #if there is no overlap
            else:
                rangesO += [r1]
        #Include the last entry
        if(rangesO[-1][1] < rangesT[-1][1]):
            rangesO += [rangesT[-1]]
        rangesT = rangesO
    return rangesO


def removeRange(ranges1, range2):
    #for a set of 'exclude wavelength ranges' in ranges1,
    #remove region in range2, so that the region in range2 is 'included'
    rangesT = sorted(ranges1, key=lambda ran: ran[0])
    rangesO = []
    for i, r1 in enumerate(rangesT):
        #if this 'exclude' range starts in the new 'include' range
        if (range2[0] <= r1[0]) and (r1[0] < range2[1]):
            #and this exclude range ends after the new include range
            if range2[1] < r1[1]:
                rangesO += [[range2[1], r1[1]]]
            #Otherwise the exclude region starts and ends in the include range
            #so skip it
        #if this 'exclude' range ends in the new 'include' range
        elif (range2[0] <= r1[1]) and (r1[1] < range2[1]):
            #and this exclude range starts before the new include range
            if r1[0] < range2[0]:
                rangesO += [[r1[0], range2[0]]]
        #If this 'exclude' ranges spans the new include range
        elif (r1[0] < range2[0]) and (range2[1] < r1[1]):
            if range2[0] < range2[1]: #sanity check
                rangesO += [[r1[0], range2[0]]]
                rangesO += [[range2[1], r1[1]]]
        #otherwise this exclude range should be outside the include range
        else:
            rangesO += [r1]
    return rangesO


#A fairly simple, fairly general tooltip
class ToolTip(object):
    """
    Create a tooltip for a given widget, using Tkinter.
    This is a fairly simple class, but it wraps text at a specified length,
    puts the tooltip below the widget, and moves the tooltip to keep it 
    within the screen limits.  
    Based on the recipe from:
    http://www.voidspace.org.uk/python/weblog/arch_d7_2006_07_01.shtml#e387
    and some discussion from: 
    https://stackoverflow.com/questions/3221956/how-do-i-display-tooltips-in-tkinter
    """
    def __init__(self, widget, text='widget info',
                 waittime = 750, wraplength = 180):
        self.waittime = waittime       #miliseconds
        self.wraplength = wraplength   #pixels
        self.widget = widget
        self.text = text
        self.widget.bind("<Enter>", self.enter)
        self.widget.bind("<Leave>", self.leave)
        self.widget.bind("<ButtonPress>", self.leave)
        self.waitid = None
        self.tw = None

    def enter(self, event=None):
        self.schedule()

    def leave(self, event=None):
        self.unschedule()
        self.hidetip()

    def schedule(self):
        self.unschedule()
        self.waitid = self.widget.after(self.waittime, self.showtip)

    def unschedule(self):
        waitid = self.waitid
        self.waitid = None
        if waitid:
            self.widget.after_cancel(waitid)

    def showtip(self, event=None):
        #Create a window showing the tooltip
        #shouldn't be necessary, but just in case remove any existing window
        self.hidetip() 
        #Make the tooltip appear outside the mouse position (below the button)
        #helps prevent accidental loops of showing and hiding the tooltip.  
        x = self.widget.winfo_pointerx() + 1
        y = self.widget.winfo_rooty() + self.widget.winfo_height() + 1
        #Creates a toplevel window to hold the tooltip
        #and remove all the standard window parts.
        self.tw = tk.Toplevel(self.widget)
        self.tw.wm_overrideredirect(True)
        label = tk.Label(self.tw, text=self.text, justify='left',
                       background="#ffffff", relief='solid', borderwidth=1,
                       wraplength = self.wraplength)
        #extra 1 pixel spacing around the text
        label.pack(ipadx=1) 
        #Check that the tooltip will fit on the screen,
        #and if it won't move the window containing it.
        if (x + label.winfo_reqwidth() + 2) > self.widget.winfo_screenwidth():
            x = self.widget.winfo_screenwidth() - (label.winfo_reqwidth() + 2)
        if (y + label.winfo_reqheight()) > self.widget.winfo_screenheight():
            y = self.widget.winfo_rooty() - label.winfo_reqheight() - 2
        #The new window seems to not draw until update_idletasks is called
        #or when this function exits and control returns to the main loop.
        #So we can change the after everything else.
        self.tw.wm_geometry("+%d+%d" % (x, y))

    def hidetip(self):
        #set self.tw to None first, in case the destroy method throws an exception
        tw = self.tw
        self.tw= None
        if tw:
            tw.destroy()


#Functions for panning and Zooming
#A little event manager thing for dealing with callbacks
class viewFuncs:
    def __init__(self, ax, canvas):
        self.ax = ax
        #Get the (first) Line2D object from this Axes object,
        #which should be the plot() of the observation.
        self.plObs = ax.get_lines()[0]
        self.canvas = canvas
        self.zoomRecActive = False
        self.lastrect = None
        return
    def autoScale(self, *event):
        fracSpace = 0.05
        xmin = np.min(self.plObs.get_xdata())
        xmax = np.max(self.plObs.get_xdata())
        ymin = np.min(self.plObs.get_ydata())
        ymax = np.max(self.plObs.get_ydata())
        xspan = xmax-xmin
        yspan = ymax-ymin
        self.ax.set_xlim(xmin-xspan*fracSpace, xmax+xspan*fracSpace)
        self.ax.set_ylim(ymin-yspan*fracSpace, ymax+yspan*fracSpace)
        self.canvas.draw()
        return
    def autoScaleY(self, *event):
        fracSpace = 0.05
        xmin, xmax = self.ax.get_xlim()
        iInViewX = (self.plObs.get_xdata() > xmin) & (self.plObs.get_xdata() < xmax)
        ymin = np.min(self.plObs.get_ydata()[iInViewX])
        ymax = np.max(self.plObs.get_ydata()[iInViewX])
        yspan = ymax-ymin
        self.ax.set_ylim(ymin-yspan*fracSpace, ymax+yspan*fracSpace)        
        self.canvas.draw()
        return
    def zoomIn(self, *event):
        fracZoom = 0.1
        xmin, xmax = self.ax.get_xlim()
        ymin, ymax = self.ax.get_ylim()
        xspan = xmax-xmin
        yspan = ymax-ymin
        modFracZoom = fracZoom/(1.+2.*fracZoom) #modify so zoom in reverses zoom out exactly
        self.ax.set_xlim(xmin+xspan*modFracZoom, xmax-xspan*modFracZoom)
        self.ax.set_ylim(ymin+yspan*modFracZoom, ymax-yspan*modFracZoom)
        self.canvas.draw()
        return
    def zoomOut(self, *event):
        fracZoom = 0.1
        xmin, xmax = self.ax.get_xlim()
        ymin, ymax = self.ax.get_ylim()
        xspan = xmax-xmin
        yspan = ymax-ymin
        self.ax.set_xlim(xmin-xspan*fracZoom, xmax+xspan*fracZoom)
        self.ax.set_ylim(ymin-yspan*fracZoom, ymax+yspan*fracZoom)
        self.canvas.draw()
        return
    def panLeft(self, *event):
        fracPan = 0.2
        xmin, xmax = self.ax.get_xlim()
        xspan = xmax-xmin
        self.ax.set_xlim(xmin-xspan*fracPan, xmax-xspan*fracPan)
        self.canvas.draw()
        return
    def panRight(self, *event):
        fracPan = 0.2
        xmin, xmax = self.ax.get_xlim()
        xspan = xmax-xmin
        self.ax.set_xlim(xmin+xspan*fracPan, xmax+xspan*fracPan)
        self.canvas.draw()
        return
    def panUp(self, *event):
        fracPan = 0.2
        ymin, ymax = self.ax.get_ylim()
        yspan = ymax-ymin
        self.ax.set_ylim(ymin+yspan*fracPan, ymax+yspan*fracPan)
        self.canvas.draw()
        return
    def panDown(self, *event):
        fracPan = 0.2
        ymin, ymax = self.ax.get_ylim()
        yspan = ymax-ymin
        self.ax.set_ylim(ymin-yspan*fracPan, ymax-yspan*fracPan)
        self.canvas.draw()
        return
    
    def zoomRec(self, *event):
        #Zoom to a rectangle selected by the user with the mouse
        if not self.zoomRecActive:
            self.zoomRecActive = True
            self.canvasWidget = self.canvas.get_tk_widget()
            self.numClick = 0
            self.x0rec, self.y0rec, self.x1rec, self.y1rec = 0., 0., 0., 0.
            self.oldBindClick = self.canvasWidget.bind('<Button-1>')
            self.canvasWidget.bind('<Button-1>', self.onClickRec)
            self.oldBindClick3 = self.canvasWidget.bind('<Button-3>')
            self.canvasWidget.bind('<Button-3>', self.onClickCancel)
            self.oldCursor = self.canvasWidget.cget('cursor')
            self.canvasWidget.config(cursor='crosshair')
        return
    def zoomRecDeactivate(self):
        self.canvasWidget.bind('<Button-1>', self.oldBindClick)
        self.canvasWidget.bind('<Button-3>', self.oldBindClick3)
        self.canvasWidget.config(cursor=self.oldCursor)
        self.zoomRecActive = False
        #when deactivating also set the plot view to the selected range
        if self.x1rec > 0:
            #get the canvas position in window pixel coordinates
            x0full, x1full = self.canvas.figure.axes[0].bbox.intervalx
            y0inv, y1inv = self.canvas.figure.axes[0].bbox.intervaly
            #and correct for matplotlib using y in the opposite direction from tk
            height = self.canvas.figure.bbox.height
            y0full = height - y1inv
            y1full = height - y0inv
            axesxrange = x1full - x0full
            axesyrange = y1full - y0full
            #get the current view range in data coordinates
            axdatax0, axdatax1, axdatay0, axdatay1 = self.canvas.figure.axes[0].axis()
            dataxrange = axdatax1 - axdatax0
            datayrange = axdatay1 - axdatay0
            #convert the selected window pixel range into data coordinates
            datax0 = (self.x0rec - x0full)/(axesxrange)*(dataxrange) + axdatax0
            datax1 = (self.x1rec - x0full)/(axesxrange)*(dataxrange) + axdatax0
            datay0 = (self.y0rec - y0full)/(axesyrange)*(-datayrange) + axdatay1
            datay1 = (self.y1rec - y0full)/(axesyrange)*(-datayrange) + axdatay1
            #set the plotted range into the selected range in data coordinates
            if abs(datax1 - datax0) > 0. and abs(datay1 - datay0) > 0.:
                self.ax.set_xlim(min(datax0, datax1), max(datax0, datax1))
                self.ax.set_ylim(min(datay0, datay1), max(datay0, datay1))
                self.canvas.draw()
        return
    def onClickRec(self, event):
        #On the first click start drawing a selection rectangle
        if self.numClick == 0:
            self.numClick += 1
            self.oldBindMove = self.canvasWidget.bind('<Motion>')
            self.canvasWidget.bind('<Motion>', self.onMoveRec)
            self.x0rec = event.x
            self.y0rec = event.y
            self.lastrect = self.canvasWidget.create_rectangle(
                self.x0rec, self.y0rec, self.x0rec+1, self.y0rec+1, width=1)
        #On the second click deactivate and finish the zoom
        else:
            self.x1rec = event.x
            self.y1rec = event.y
            if self.lastrect is not None:
                self.canvasWidget.delete(self.lastrect)
            self.canvasWidget.bind('<Motion>', self.oldBindMove)
            self.zoomRecDeactivate()
        return
    def onClickCancel(self, event):
        #cancel the zoom to rectangle, usually in right click
        self.canvasWidget.delete(self.lastrect)
        if self.numClick > 0:
            self.canvasWidget.bind('<Motion>', self.oldBindMove)
        self.zoomRecDeactivate()
        return
    def onMoveRec(self, event):
        #delete and redraw the rectangle when the mouse moves
        x1 = event.x
        y1 = event.y
        self.canvasWidget.delete(self.lastrect)
        self.lastrect = self.canvasWidget.create_rectangle(
            self.x0rec, self.y0rec, x1, y1, width=1)
        return


#A base class for uiExcludeRange and uiIncludeRange
class uiRangeBase:
    def __init__(self, canvas, mask, pltMaskU, pltMaskN, pltMaskF,
                 excludeRanges, fitDepthFlags):
        self.canvas = canvas
        self.canvasWidget = canvas.get_tk_widget()
        self.mask = mask
        self.pltMaskU = pltMaskU
        self.pltMaskN = pltMaskN
        self.pltMaskF = pltMaskF
        self.excludeRanges = excludeRanges
        self.fitDepthFlags = fitDepthFlags
        self.active = False
        self.rangeSelect = rangeSelect(canvas, self)

        #Set up a style for the 'actively selecting a range' button state
        #Seems like I need to set the base style changes first,
        #then use map to set state specific differences,
        #since map doesn't provide easy access to being in
        #not any of the special states (only not a specific state)
        self.style = ttk.Style()
        self.style.configure('ActRange.TButton', relief='sunken',
                        background='#80ccff')
        self.style.map('ActRange.TButton',
                  relief=[('active','sunken')],
                  background=[('pressed','#eeeeee'),('active','#80ccff')])
        #note: order matters for map here, 
        #when the button is 'active' (mouse over) and also being pressed
        #The scope of the style names seems to be pretty wide...

    def linkButton(self, button, otherRangeList):
        self.button = button
        self.otherRangeList = otherRangeList
    def runSpanSelect(self):
        #First turn off the other linked buttons (only one mode active!)
        for otherRange in self.otherRangeList:
            if (otherRange.active == True):
                otherRange.deactivate()
        self.active = not self.active
        if self.active:
            self.button.configure(style='ActRange.TButton')
            #get a string of tk info that we can use later to restore the current bindings
            self.oldBindClick = self.canvasWidget.bind('<Button-1>')
            #override the current binding
            self.fidClick = self.canvasWidget.bind('<Button-1>',
                                                   self.rangeSelect.onClick)
            self.oldBindClick3 = self.canvasWidget.bind('<Button-3>')
            self.canvasWidget.bind('<Button-3>', self.rangeSelect.cancel)
        else:
            self.deactivate()
    def deactivate(self):
        #stop span selector...
        self.rangeSelect.deactivate()
        self.button.configure(style='TButton')
        #reset to the previous bindings once we are done
        #(Safer to bind to something different, since unbind seems to have poorly defined behaviour)
        self.canvasWidget.bind('<Button-1>', self.oldBindClick)
        self.canvasWidget.bind('<Button-3>', self.oldBindClick3)
        self.active = False


class uiIncludeRange(uiRangeBase):
    def selectedWl(self, minval, maxval):
        #After a region has been selected, update the mask and 'exclude regions'
        self.mask.iuse[(self.mask.wl >= minval) & (self.mask.wl <= maxval)] = 1
        self.excludeRanges[:] = removeRange(self.excludeRanges, [minval,maxval])
        
        updateLinePlots(self.mask, self.fitDepthFlags,
                        self.pltMaskU, self.pltMaskN, self.pltMaskF)
        self.canvas.draw()

        
class uiExcludeRange(uiRangeBase):
    def selectedWl(self, minval, maxval):
        #After a region has been selected, update the mask and 'exclude regions'
        self.mask.iuse[(self.mask.wl >= minval) & (self.mask.wl <= maxval)] = 0
        self.excludeRanges[:] = combineRanges(self.excludeRanges,
                                              [[minval,maxval]])
        
        updateLinePlots(self.mask, self.fitDepthFlags,
                        self.pltMaskU, self.pltMaskN, self.pltMaskF)
        self.canvas.draw()


def updateLinePlots(mask, fitDepthFlags, pltMaskU, pltMaskN, pltMaskF):
    #Update the plot of lines from the mask,
    #with plots for lines used, lines excluded,
    #and lines with depths to fit.
    maskUsed = mask[mask.iuse == 1]
    maskNot = mask[mask.iuse == 0]
    #update the data in the plot of lines
    segments = []
    for miniM in maskUsed:
        segments += [np.array([[miniM.wl, 1.0-miniM.depth],
                              [miniM.wl, 1.0]])]
    pltMaskU.set_segments(segments)
    segments = []
    for miniM in maskNot:
        segments += [np.array([[miniM.wl, 1.0-miniM.depth],
                              [miniM.wl, 1.0]])]
    pltMaskN.set_segments(segments)
    #including the plot of lines highlighted for depth fitting
    maskFit = mask[(fitDepthFlags == 1) & (mask.iuse == 1)]
    segments = []
    for miniM in maskFit:
        segments += [np.array([[miniM.wl, 1.0-miniM.depth],
                               [miniM.wl, 1.0]])]
    pltMaskF.set_segments(segments)
    return


class uiSelectFitDepth(uiRangeBase):
    def selectedWl(self, minval, maxval):
        #After a region has been selected, update selected lines
        #update the data in the plot of lines
        fitDF = self.fitDepthFlags
        fitDF[(self.mask.wl >= minval) & (self.mask.wl <= maxval)] = 1
        maskFit = self.mask[(fitDF == 1) & (self.mask.iuse == 1)]
        segments = []
        for miniM in maskFit:
            segments += [np.array([[miniM.wl, 1.0-miniM.depth],
                                   [miniM.wl, 1.0]])]
        self.pltMaskF.set_segments(segments)
        self.canvas.draw()


class uiUnselectFitDepth(uiRangeBase):
    def selectedWl(self, minval, maxval):
        #After a region has been selected, update selected lines
        #update the data in the plot of lines
        fitDepthFlags = self.fitDepthFlags
        fitDepthFlags[(self.mask.wl >= minval) & (self.mask.wl <= maxval)] = 0
        maskFit = self.mask[(fitDepthFlags == 1) & (self.mask.iuse == 1)]
        segments = []
        for miniM in maskFit:
            segments += [np.array([[miniM.wl, 1.0-miniM.depth],
                                   [miniM.wl, 1.0]])]
        self.pltMaskF.set_segments(segments)
        self.canvas.draw()


#Handle drawing a range selection on the screen, and getting the range selected
class rangeSelect:
    def __init__(self, canvas, parentUIRange):
        self.canvas = canvas
        self.canvasWidget = canvas.get_tk_widget()
        self.parentUIRange = parentUIRange
        self.bStartSelect = True
        self.lastrect = None
        
    def onClick(self, event):
        #Tkinter uses units from the top left, while matplotlib provides units from the bottom left.
        #So we need to flip matplotlilb's provided values (even in screen space)
        x0full, x1full = self.canvas.figure.axes[0].bbox.intervalx
        y0full, y1full = self.canvas.figure.axes[0].bbox.intervaly
        height = self.canvas.figure.bbox.height
        y0mod = height - y1full
        y1mod = height - y0full
        
        if self.bStartSelect:  #For the 1st click
            self.x0 = x0 = event.x
            self.y0 = y0 = event.y
            
            #drawn then update a rectangle as the mouse moves
            self.bStartSelect = False
            self.lastrect = self.canvasWidget.create_rectangle(x0, y0mod, x0, y1mod)
            self.oldBindMove = self.canvasWidget.bind('<Motion>')
            self.funcid = self.canvasWidget.bind('<Motion>', self.onMovePointer)
        else:
            self.x1 = event.x
            self.y1 = event.y

            #Get the cursor position in plotted data coordinates
            #(Testing this against matplotlib's toolbar it seems accurate to a pixel level.)
            axesdatax0, axesdatax1, axesdatay0, axesdatay1 = self.canvas.figure.axes[0].axis()
            axesxrange = x1full - x0full
            dataxrange = axesdatax1 - axesdatax0
            self.datax0 = (self.x0 - x0full)/(axesxrange)*(dataxrange) + axesdatax0
            self.datax1 = (self.x1 - x0full)/(axesxrange)*(dataxrange) + axesdatax0
            
            #Call the creating object's function for processing this range.
            #This is a bit recursive, so there may be a small possibility
            #to leak memory or odd behaviour?
            self.parentUIRange.selectedWl(min(self.datax0, self.datax1),
                                          max(self.datax0, self.datax1))
            self.deactivate()

    def cancel(self, *event):
            self.deactivate()
            
    def deactivate(self):
        #if there is an active selection region (not starting to select) undo it
        if not self.bStartSelect:
            self.bStartSelect = True
            self.canvasWidget.delete(self.lastrect)
            #Bind can take a string of tk binding info,
            #and bind returns such a string if called with just an event name,
            #so we can use that to restore a previous binding.
            self.funcid = self.canvasWidget.bind('<Motion>', self.oldBindMove)

    def onMovePointer(self, event):
        #Delete and redraw the rectangle when the mouse moves
        x1 = event.x
        y1 = event.y
        self.canvasWidget.delete(self.lastrect)
        #Get the full vertical range of the plot for the rectangle
        y0full, y1full = self.canvas.figure.axes[0].bbox.intervaly
        #Change from matplotlib to tk's y-axis direction
        height = self.canvas.figure.bbox.height
        y0mod = height - y1full
        y1mod = height - y0full
        self.lastrect = self.canvasWidget.create_rectangle(self.x0, y0mod, x1, y1mod, dash=[5,5], width=2)
        #self.lastrect = self.canvasWidget.create_rectangle(self.x0, y0mod, x1, y1mod, dash=[3,3])


class saveRanges:
    #mini manager for the button that saves the set of exclude ranges to file
    def __init__(self, fname, ranges):
        self.fname = fname
        self.ranges = ranges
    def saveToFile(self):
        #make an ExcludeMaskRegions object
        wlStarts = np.array([ran[0] for ran in self.ranges])
        wlEnds = np.array([ran[1] for ran in self.ranges])
        labels = np.array(['cleanMaskUI']*len(self.ranges),dtype=object)
        regions = maskTools.ExcludeMaskRegions(wlStarts, wlEnds, labels)
        #Save the set of exclude ranges to a file
        regions.save(self.fname)
        return


class fitDepths:
    #mini manager to run the fitting of line depths
    def __init__(self, mask, obs, lsdp, lsdProf, fitDepthFlags, root,
                 canvas, pltMaskU, pltMaskN, pltMaskF):
        self.mask = mask
        self.obs = obs
        self.lsdp = lsdp
        self.lsdProf = lsdProf
        self.fitDepthFlags = fitDepthFlags
        self.root = root
        self.canvas = canvas
        self.pltMaskU = pltMaskU
        self.pltMaskN = pltMaskN
        self.pltMaskF = pltMaskF
        
    def runFit(self):
        #Fit line depths using a linear least squares method
        #and then update the plots of lines from the mask.
        #This automatically removes degenerate or near-degenerate
        #lines (i.e. very near in wavelength) before fitting depths.

        #First set a 'wait' cursor
        oldCursor = self.root.cget('cursor')
        self.root.config(cursor='watch')
        self.root.update()

        #Get the lines that are flagged to be fit (and used)
        indUse1 = np.nonzero((self.fitDepthFlags == 1) & (self.mask.iuse == 1))
        useMask1 = self.mask[indUse1]
        if len(useMask1) > 0:
            #Get the reference LSD profile
            prof = self.lsdProf
            #Remove degenerate (or nearly) lines
            pixVel = prof.vel[1]-prof.vel[0]
            removePoorLines(useMask1, pixVel, fracPix = 3.0, sumDepths=False)
            self.mask.iuse[indUse1] = useMask1.iuse
            #and save the good lines in the mask for fitting
            indUse2 = np.nonzero(useMask1.iuse == 1)
            useMask2 = useMask1[indUse2]
            
            normDepth = self.lsdp.normDepth
            #Fit in two steps, and save the result back to the mask
            MD = buildMD(self.obs, useMask2, prof, normDepth)
            weights = linlsqDepths(MD, useMask2, self.obs)
            self.mask.depth[indUse1[0][indUse2]] = weights
            #update the plot
            updateLinePlots(self.mask, self.fitDepthFlags,
                            self.pltMaskU, self.pltMaskN, self.pltMaskF)
            self.canvas.draw()

        #Return the cursor to normal
        self.root.config(cursor=oldCursor)
        return

class undoDepths:
    #mini manager to undo the fitting of line depths
    def __init__(self, mask, fitDepthFlags, canvas,
                 pltMaskU, pltMaskN, pltMaskF):
        self.mask = mask
        self.fitDepthFlags = fitDepthFlags
        self.canvas = canvas
        self.pltMaskU = pltMaskU
        self.pltMaskN = pltMaskN
        self.pltMaskF = pltMaskF
        self.oldDepths = np.zeros_like(mask.depth)
        self.oldDepths[:] = mask.depth[:] #make a copy instead of a reference

    def undo(self):
        fUse = (self.fitDepthFlags == 1) & (self.mask.iuse == 1)
        self.mask.depth[fUse] = self.oldDepths[fUse]
        updateLinePlots(self.mask, self.fitDepthFlags,
                        self.pltMaskU, self.pltMaskN, self.pltMaskF)
        self.canvas.draw()
        return


#Make a window for modifying some LSD calculation parameters,
#includes associated functions for handling the UI
class modifyLSDpar:
    def __init__(self, parent, lsdp):
        self.parent = parent
        self.lsdp = lsdp
        self.active = False
        self.win = None

        self.varSaveMod = tk.IntVar()
        if self.lsdp.outModelName is None:
            self.varSaveMod.set(0)
        else:
            self.varSaveMod.set(1)
    
    def closeWindow(self, *event):
        #Update any input information
        self.setVelStart()
        self.setVelEnd()
        self.setVelPix()
        self.setNDepth()
        self.setNWav()
        self.setNLande()
        self.setModName()
        self.setProfName()
        
        self.active = False
        try:
            self.win.destroy()
        except:
            pass
        return
    
    def openWindow(self, *event):
        #If the window is already open do nothing
        if self.active:
            return
        self.active = True
        
        self.win = tk.Toplevel(self.parent)
        self.win.title("Set LSD parameters")
        #Associate this window with its parent,
        #and keep this window out of the window manager.
        self.win.transient(self.parent)
        #Send keyboard and mouse events only to this window,
        #mostly deactivating other windows.
        self.win.grab_set()
        #Set what happens when the window manager asks to close the window.
        self.win.protocol("WM_DELETE_WINDOW", self.closeWindow)

        wframe = ttk.Labelframe(self.win, text='Set parameters for LSD profile',
                                padding="10 10 10 10")
        wframe.pack(fill=tk.BOTH, expand=1, padx=4, pady=4)

        #Modify the velocity grid for the LSD profile
        labVelStart = ttk.Label(wframe, text='start vel', justify='right')
        self.txtVelStart = tk.StringVar()
        self.txtVelStart.set('{:.1f}'.format(self.lsdp.velStart))
        entVelStart = ttk.Entry(master=wframe, textvariable=self.txtVelStart,
                                width=8, validate='focusout',
                                validatecommand=self.setVelStart)
        entVelStart.bind('<Key-Return>', self.setVelStart)
        ToolTip(entVelStart, 'Set the starting velocity for the LSD profile (km/s)')
        labVelEnd = ttk.Label(wframe, text=' end vel', justify='right')
        self.txtVelEnd = tk.StringVar()
        self.txtVelEnd.set('{:.1f}'.format(self.lsdp.velEnd))
        entVelEnd = ttk.Entry(master=wframe, textvariable=self.txtVelEnd,
                              width=8, validate='focusout',
                              validatecommand=self.setVelEnd)
        entVelEnd.bind('<Key-Return>', self.setVelEnd)
        ToolTip(entVelEnd, 'Set the ending velocity for the LSD profile (km/s)')
        labVelPix = ttk.Label(wframe, text=' pixel size', justify='right')
        self.txtVelPix = tk.StringVar()
        self.txtVelPix.set('{:.2f}'.format(self.lsdp.velPixel))
        entVelPix = ttk.Entry(master=wframe, textvariable=self.txtVelPix,
                              width=8, validate='focusout',
                              validatecommand=self.setVelPix)
        entVelPix.bind('<Key-Return>', self.setVelPix)
        ToolTip(entVelPix, 'Set the pixel size of the LSD profile in km/s')

        #Modify LSD normalization values
        labNDepth = ttk.Label(wframe, text='norm. depth', justify='right')
        self.txtNDepth = tk.StringVar()
        self.txtNDepth.set('{:.2f}'.format(self.lsdp.normDepth))
        entNDepth = ttk.Entry(master=wframe, textvariable=self.txtNDepth,
                              width=8, validate='focusout',
                            validatecommand=self.setNDepth)
        entNDepth.bind('<Key-Return>', self.setNDepth)
        ToolTip(entNDepth, 'Set the normalizing depth for the LSD mask/profile')
        labNWav = ttk.Label(wframe, text=' norm. wavelength',
                            justify='right')
        self.txtNWav = tk.StringVar()
        self.txtNWav.set('{:.1f}'.format(self.lsdp.normWave))
        entNWav = ttk.Entry(master=wframe, textvariable=self.txtNWav,
                            width=8, validate='focusout',
                            validatecommand=self.setNWav)
        entNWav.bind('<Key-Return>', self.setNWav)
        ToolTip(entNWav, 'Set the normalizing wavelength for the LSD mask/profile (usually nm)')
        labNLande = ttk.Label(wframe, text=' norm. Lande',
                              justify='right')
        self.txtNLande = tk.StringVar()
        self.txtNLande.set('{:.2f}'.format(self.lsdp.normLande))
        entNLande = ttk.Entry(master=wframe, textvariable=self.txtNLande,
                              width=8, validate='focusout',
                            validatecommand=self.setNLande)
        entNLande.bind('<Key-Return>', self.setNLande)
        ToolTip(entNLande, 'Set the normalizing Landé factor for the LSD mask/profile')

        #remove closely spaced lines in the LSD calculation?
        self.varRemoveClose = tk.IntVar()
        self.varRemoveClose.set(int(self.lsdp.trimMask))
        checkRemoveClose = ttk.Checkbutton(wframe,
                                           text='remove closely\nspaced lines',
                                           variable=self.varRemoveClose,
                                           command=self.setRemoveClose)
        ToolTip(checkRemoveClose, "Remove very closely spaced lines in the mask when calculating the LSD profile")

        #Optionally save the model LSD spectrum
        checkSaveMod = ttk.Checkbutton(wframe,
                                       text='save model\nspectrum',
                                       variable=self.varSaveMod,
                                       command=self.setSaveMod)
        ToolTip(checkSaveMod, "Save a copy of the model LSD spectrum to a file?")
        labModName = ttk.Label(wframe, text='model\nfile name',
                               justify='right')
        self.txtModName = tk.StringVar()
        if self.lsdp.outModelName is None:
            self.txtModName.set('')
        else:
            self.txtModName.set('{:}'.format(self.lsdp.outModelName))
        entModName = ttk.Entry(master=wframe,
                               textvariable=self.txtModName, width=25,
                               validate='focusout',
                               validatecommand=self.setModName)
        entModName.bind('<Key-Return>', self.setModName)
        ToolTip(entModName, 'File name for the output model LSD spectrum')
        
        #Set the output LSD profile file name, and optionally plot it.
        labProfName = ttk.Label(wframe, text='profile\nfile name',
                                justify='right')
        self.txtProfName = tk.StringVar()
        self.txtProfName.set('{:}'.format(self.lsdp.outName))
        entProfName = ttk.Entry(master=wframe,
                                textvariable=self.txtProfName, width=25, 
                                validate='focusout',
                                validatecommand=self.setProfName)
        entProfName.bind('<Key-Return>', self.setProfName)
        ToolTip(entProfName, 'File name for the output LSD profile')

        self.varPltProf = tk.IntVar()
        self.varPltProf.set(int(self.lsdp.plotLSD))
        checkPltProf = ttk.Checkbutton(wframe, text='plot\nprofile',
                                       variable=self.varPltProf,
                                       command=self.setPlotProf)
        ToolTip(checkPltProf, "Plot the LSD profile in a new window when updated?")

        #Lay out the widgits in the window
        labVelStart.grid(row=0, column=0, sticky=tk.E, pady=4, padx=4)
        entVelStart.grid(row=0, column=1, sticky=tk.W, pady=4)
        labVelEnd.grid(row=1, column=0, sticky=tk.E, pady=4, padx=4)
        entVelEnd.grid(row=1, column=1, sticky=tk.W, pady=4)
        labVelPix.grid(row=2, column=0, sticky=tk.E, pady=4, padx=4)
        entVelPix.grid(row=2, column=1, sticky=tk.W, pady=4)

        labNDepth.grid(row=0, column=2, sticky=tk.E, pady=4, padx=4)
        entNDepth.grid(row=0, column=3, sticky=tk.W, pady=4)
        labNWav.grid(row=1, column=2, sticky=tk.E, pady=4, padx=4)
        entNWav.grid(row=1, column=3, sticky=tk.W, pady=4)
        labNLande.grid(row=2, column=2, sticky=tk.E, pady=4, padx=4)
        entNLande.grid(row=2, column=3, sticky=tk.W, pady=4)

        checkRemoveClose.grid(row=3, column=0, columnspan=4, padx=4, pady=4)

        checkSaveMod.grid(row=4, column=0, columnspan=1, pady=4)
        labModName.grid(row=4, column=1, sticky=tk.E, pady=4, padx=4)
        entModName.grid(row=4, column=2, columnspan=2, sticky=tk.W, pady=4)

        labProfName.grid(row=5, column=1, sticky=tk.E, pady=4, padx=4)
        entProfName.grid(row=5, column=2, columnspan=3, sticky=tk.W, pady=4)
        checkPltProf.grid(row=6, column=0, columnspan=4, pady=4, padx=4)

        wframe.columnconfigure(0, weight=1)
        wframe.columnconfigure(1, weight=1)
        wframe.columnconfigure(2, weight=1)
        wframe.columnconfigure(3, weight=1)
        #wframe.columnconfigure(4, weight=1)
        wframe.rowconfigure(0, weight=1)
        wframe.rowconfigure(1, weight=1)
        wframe.rowconfigure(2, weight=1)
        wframe.rowconfigure(3, weight=1)
        wframe.rowconfigure(4, weight=1)
        wframe.rowconfigure(5, weight=1)
        wframe.rowconfigure(6, weight=1)
        
        #Enter a local event loop, returning only when the given window
        #is destroyed.
        self.win.wait_window(self.win)

    def setVelStart(self, *event):
        try:  #Make sure this text is a number
            val = float(self.txtVelStart.get())
        except ValueError:
            self.txtVelStart.set('{:.1f}'.format(self.lsdp.velStart))
            return False
        if val <= -1000.:  #and make sure the value is sensible
            val = -999.0
            self.txtVelStart.set('{:.1f}'.format(val))
        elif (val > self.lsdp.velEnd - 10.):
            val = self.lsdp.velEnd - 10.
            self.txtVelStart.set('{:.1f}'.format(val))
        self.lsdp.velStart = val
        return True
    def setVelEnd(self, *event):
        try:  #Make sure this text is a number
            val = float(self.txtVelEnd.get())
        except ValueError:
            self.txtVelEnd.set('{:.1f}'.format(self.lsdp.velEnd))
            return False
        if val >= 1000.:  #and make sure the value is sensible
            val = 999.0
            self.txtVelEnd.set('{:.1f}'.format(val))
        elif (val < self.lsdp.velStart + 10.):
            val = self.lsdp.velStart + 10.
            self.txtVelEnd.set('{:.1f}'.format(val))
        self.lsdp.velEnd = val
        return True
    def setVelPix(self, *event):
        try:  #Make sure this text is a number
            val = float(self.txtVelPix.get())
        except ValueError:
            self.txtVelPix.set('{:.2f}'.format(self.lsdp.velPixel))
            return False
        if val <= 0:
            self.txtVelPix.set('{:.2f}'.format(self.lsdp.velPixel))
            return False
        elif val < 0.1:
            val = 0.1
            self.txtVelPix.set('{:.2f}'.format(val))
        elif val > (self.lsdp.velEnd - self.lsdp.velStart)/8.0:
            val = (self.lsdp.velEnd - self.lsdp.velStart)/8.0
            self.txtVelPix.set('{:.2f}'.format(val))
        self.lsdp.velPixel = val
        return True
    def setNDepth(self, *event):
        try:
            val = float(self.txtNDepth.get())
        except ValueError:
            self.txtNDepth.set('{:.2f}'.format(self.lsdp.normDepth))
            return False
        if val <= 0:
            self.txtNDepth.set('{:.2f}'.format(self.lsdp.normDepth))
            return False
        elif val > 1.0:
            val = 1.0
            self.txtNDepth.set('{:.2f}'.format(val))
        self.lsdp.normDepth = val
        return True
    def setNWav(self, *event):
        try:
            val = float(self.txtNWav.get())
        except ValueError:
            self.txtNWav.set('{:.1f}'.format(self.lsdp.normWave))
            return False
        if val <= 0:
            self.txtNWav.set('{:.1f}'.format(self.lsdp.normWave))
            return False
        elif val > 100000.0:
            val = 100000.0
            self.txtNWav.set('{:.1f}'.format(val))
        self.lsdp.normWave = val
        return True
    def setNLande(self, *event):
        try:
            val = float(self.txtNLande.get())
        except ValueError:
            self.txtNLande.set('{:.2f}'.format(self.lsdp.normLande))
            return False
        if val <= 0:
            self.txtNLande.set('{:.2f}'.format(self.lsdp.normLande))
            return False
        elif val > 10.0:
            val = 10.0
            self.txtNLande.set('{:.2f}'.format(val))
        self.lsdp.normLande = val
        return True

    def setRemoveClose(self, *event):
        val = self.varRemoveClose.get()
        self.lsdp.trimMask = bool(val)
        return
    
    def setSaveMod(self, *event):
        val = self.varSaveMod.get()
        if val == 0:
            self.lsdp.outModelName = None
        else:
            self.lsdp.outModelName = self.txtModName.get()
        return
    def setModName(self, *event):
        flag = self.varSaveMod.get()
        if flag != 0:
            self.lsdp.outModelName = self.txtModName.get()
        else:
            self.lsdp.outModelName = None
        return
    def setProfName(self, *event):
        val = self.txtProfName.get()
        if val != '':
            self.lsdp.outName = val
        return
    def setPlotProf(self, *event):
        val = self.varPltProf.get()
        if val == 0:
            self.lsdp.plotLSD = False
        else:
            self.lsdp.plotLSD = True
        return


class updateLSD:
    def __init__(self, canvas, root, mask, lsdp, lsdProf, pltModelI):
        self.canvas = canvas
        self.root = root
        self.mask = mask
        self.lsdp = lsdp
        self.lsdProf = lsdProf
        self.pltModelI = pltModelI
    def rerunLSD(self):
        #Recalculate the LSD profile and update the plot
        #First set a 'wait' cursor
        oldCursor = self.root.cget('cursor')
        self.root.config(cursor='watch')
        self.root.update()
        #Run LSD
        lsdp = self.lsdp
        self.mask.save(lsdp.mask)
        lsdProf, modelSpec = profileLSD.run_lsdpy(obs=lsdp.obs, mask=lsdp.mask,
            outLSDName=lsdp.outName, velStart=lsdp.velStart, velEnd=lsdp.velEnd,
            velPixel=lsdp.velPixel, normDepth=lsdp.normDepth,
            normLande=lsdp.normLande, normWave=lsdp.normWave,
            removeContPol=lsdp.removeContPol, trimMask=lsdp.trimMask,
            sigmaClipIter=lsdp.sigmaClipIter, sigmaClip=lsdp.sigmaClip,
            outModelName=lsdp.outModelName,
            plotLSD=lsdp.plotLSD, outPlotLSDName=lsdp.outPlotLSDName)
        #update the current LSD object, rather than replacing it
        #(since fitDepths needs to retain a reference to the same object)
        #self.lsdProf[:] = lsdProf[:]
        self.lsdProf.vel = lsdProf.vel
        self.lsdProf.specI = lsdProf.specI
        self.lsdProf.specSigI = lsdProf.specSigI
        self.lsdProf.specV = lsdProf.specV
        self.lsdProf.specSigV = lsdProf.specSigV
        self.lsdProf.specN1 = lsdProf.specN1
        self.lsdProf.specSigN1 = lsdProf.specSigN1
        self.lsdProf.specN2 = lsdProf.specN2
        self.lsdProf.specSigN2 = lsdProf.specSigN2
        self.lsdProf.npix = lsdProf.npix
        self.lsdProf.numParam = lsdProf.numParam
        self.lsdProf.header = lsdProf.header
        #Return the cursor to normal
        self.root.config(cursor=oldCursor)
        
        #re-draw points used in the fit
        for dline in self.pltModelI:
            dline.set_data(modelSpec.wl, modelSpec.specI)
        self.canvas.draw()
        return


def buildMD(obs, mask, prof, normDepth):
    #Set up an matrix that that can be multiplied with a vector of line weights
    #to produce a mode LSD spectrum, array dimensions are nObsPts x nMaskLines
    #This has columns for each line in the mask that each essentially 
    #contain the LSD profile, with the profile shifted to be centred 
    #on the line's wavelength, and interpolated onto the wavelengths of 
    #the observation (the row pixels).  The the dot product of this
    #matrix with the vector of line mask weights then scales each copy of 
    #the LSD profile by that line's weight, and then sums over all lines.
    #Thus this dot product the LSD model spectrum.
    cvel = 2.99792458e5 #c in km/s

    #calculate wavelengths for the profile at each line in the mask here, since it is reusable
    wlProfA = np.outer(prof.vel/cvel, mask.wl) + np.tile(mask.wl, (prof.npix,1))  #(prof.npix, mask.wl.shape)

    #We can use a sparse "List of Lists" matrix to efficiently build the desired
    #matrix, but we have to start with the transpose of what we want for
    #efficient access when adding (non-zero) entries to the lil_matrix.
    Dlil = scipy.sparse.lil_matrix((mask.wl.shape[0], obs.wl.shape[0]))

    for l in range(mask.wl.shape[0]):
        #Get observation points in range of this line in the mask
        iObsRange = np.where( (wlProfA[0,l] < obs.wl[:]) & (wlProfA[-1,l] > obs.wl[:]) )
        #Interpolate the LSD profile for this line onto the observation
        interpProfI = np.interp(obs.wl[iObsRange], wlProfA[:,l], 1.-prof.specI)

        Dlil[l, iObsRange] = interpProfI / normDepth

    #Convert the matrix into a more efficient sparse form (csr_matrix) for matrix math
    Dout = Dlil.tocsr().T
    
    ##Output the model spectrum, for testing purposes only.
    #weightI = mask.depth/normDepth
    #specT = 1.-Dout.dot(weightI)
    #fOutPre = open('outSpecBeforeTweak.dat', 'w')
    #for i in range(obs.wl.shape[0]):
    #    fOutPre.write('{:} {:}\n'.format(obs.wl[i], specT[i]))
    #fOutPre.close()    
    return Dout

def linlsqDepths(maskMD, mask, obs):
    #Perform the linear least squares fit to the observation
    #to find the optimal line depths for the LSD mask.
    #This follow the same logic as the LSD profile calculation in LSDpy,
    #or most other linear least squares calculations. Here the model is 
    #Y = D.W, for the vector of line depths W, and 2D matrix M.  For that 
    #model with observations Yo, with S a diagonal matrix of 1/sigma
    # chi^2 = (Yo - D.W)^T.(S^2).(Yo - D.W)
    #Taking the derivative of chi^2 wrt W and setting it equal to 0 
    #to find the minimum in chi^2 gives
    # 0 = (-D)^T.(S^2).(Yo - D.W)*2
    # W = (D^T.(S^2).D)^(-1).D^T.(S^2).Yo
    #so we mostly just need to invert the matrix (D^T.(S^2).D)
    #and also build D^T.(S^2).Yo

    #We are working in shifted units with the continuum at 0 and lines +ve
    obsI = 1. - obs.specI
    
    #Use some sparse matrices here to be efficient
    #The dot products for alpha and beta can be slower than matrix inversion
    #Use similar choices to LSDpy for sparse routines
    sparseS2 = scipy.sparse.diags(obs.specSig**(-2), offsets=0)
    MD = scipy.sparse.csr_matrix(maskMD) #probably already a sparse csr
    
    beta = MD.T.dot(sparseS2.dot(obsI))
    alpha = MD.T.dot(sparseS2.dot(MD))

    covar = np.linalg.inv(alpha.toarray())
    lineWeights = np.dot(covar, beta)
    #lineWeightErrs = np.sqrt(np.diag(covar)) #if errors are needed
    return lineWeights


def removePoorLines(mask, pixVel, fracPix = 1.0, sumDepths=True):
    #Remove nearly degenerate lines from the mask.
    #Reject lines separated by less than fracPix of an LSD (velocity) pixel.
    #(Based on LSDpy)
    cvel = 2.99792458e5 #c in km/s
    depthCutoff = 0.6
    nTrimmed = 0
    for l in range(1,mask.wl.shape[0]):
        #This loop is relatively inefficient but handles unsorted line masks
        #and lines with multiple bad blends.
        if mask.iuse[l] == 1:
            deltas = np.abs(mask.wl[l] - mask.wl)/mask.wl[l]*cvel
            iClose = np.nonzero((deltas < pixVel*fracPix) & (mask.iuse == 1))[0]
            if iClose.shape[0] > 1:
                #If other lines are too close to the current line
                mask.iuse[iClose] = 0
                deepestLine = np.argmax(mask.depth[iClose])
                mask.iuse[iClose[deepestLine]] = 1
                
                #If we want to sum line depths, limit the maximum depth lines 
                #can sum to, as a very rough approximation for saturation.
                if sumDepths:
                    summedDepth = np.sum(mask.depth[iClose])
                    if summedDepth < depthCutoff:
                        mask.depth[iClose[deepestLine]] = summedDepth
                    else:
                        mask.depth[iClose[deepestLine]] = max(depthCutoff,
                                                mask.depth[iClose[deepestLine]])
                nTrimmed += np.count_nonzero(iClose) - 1
    if nTrimmed > 0:
        print('Modified line mask, removed {:d} too closely spaced lines'.format(nTrimmed))

    ##If one wanted a shorter mask
    #mask = mask.prune()
    return


class lsdParams:
    def __init__(self, obs, mask, outName='',
                 velStart=-200., velEnd=200., velPixel=None,
                 normDepth=0.2, normLande=1.2, normWave=500.,
                 removeContPol=True, trimMask=False, 
                 sigmaClipIter=0, sigmaClip=500.,
                 outModelName=None,
                 plotLSD=False, outPlotLSDName=None):
        """
        A simple data structure to hold input parameters for LSD calculations,
        and to set some sensible defaults.
        """
        self.obs = obs
        self.mask = mask
        self.outName = outName
        self.velStart = velStart
        self.velEnd = velEnd
        self.normDepth = normDepth
        self.normLande = normLande
        self.normWave = normWave
        self.removeContPol = removeContPol
        self.trimMask = trimMask
        self.sigmaClipIter = sigmaClipIter
        self.sigmaClip = sigmaClip
        self.outModelName = outModelName
        self.plotLSD = plotLSD
        self.outPlotLSDName = outPlotLSDName
        
        cvel = 2.99792458e5 #c in km/s
        if velPixel == None:
            #Get average observed wavelength spacing
            obsSp = obsSpec.read_spectrum(self.obs)
            wlStep = obsSp.wl[1:] - obsSp.wl[:-1]
            medianVelPix = np.median(wlStep/obsSp.wl[:-1]*cvel)
            self.velPixel = round(medianVelPix, ndigits=2)
        else:
            self.velPixel = velPixel



"""
Tools for generating and manipulating line masks for LSD.
Includes creating a line mask from a VALD line list, cleaning problem 
lines from the mask, managing regions to exclude from the mask,
and interacively cleaning and tweaking.
"""

import numpy as np
import warnings
#from . import lineList as lineListLib #(moved inside some functions)

###################################
###################################

class Mask:
    """
    The data for the LSD line Mask.

    This usually contains the arrays: 

    * wl - wavelengths of lines
    * element - the element+ion code for the line
    * depth - the depth of the line
    * lande - the effective Lande factor of the line
    * iuse - integer flag for whether the line is used
    """
    def __init__(self, wl, element, depth, excite, lande, iuse):
        """
        Generate a Mask object
        """
        self.wl = wl
        self.element = element
        self.depth = depth
        self.excite = excite
        self.lande = lande
        self.iuse = iuse

    def __getitem__(self, key):
        """
        Returns a Mask object with only the values at the specified index(s).

        :param key: the index or slice being checked
        :rtype: Mask
        """
        wl_s = self.wl[key]
        element_s = self.element[key]
        depth_s = self.depth[key]
        excite_s = self.excite[key]
        lande_s = self.lande[key]
        iuse_s = self.iuse[key]
        return Mask(wl_s, element_s, depth_s, excite_s, lande_s, iuse_s)

    def __setitem__(self, key, newval):
        """
        Sets all values of the mask at the specified location
        equal to the input mask's values.

        :param key: the index or slice being overwritten
        :param newval: Mask object used to replace the overwritten values
        """
        if not(isinstance(newval, Mask)):
            raise TypeError()
        else:
            self.wl[key] = newval.wl
            self.element[key] = newval.element
            self.depth[key] = newval.depth
            self.excite[key] = newval.excite
            self.lande[key] = newval.lande
            self.iuse[key] = newval.iuse

    def __len__(self):
        """
        Returns the number of lines in the mask.
        """
        return len(self.wl)
    
    def prune(self):
        """
        Return a Mask object with unused lines removed from the Mask.
        
        Remove lines if iuse index is set to 0,
        restricting the Mask to only lines used in LSD.
        
        :rtype: Mask
        """
        trimmedMask = self[self.iuse != 0]
        return trimmedMask
        
    def get_weights(self, normDepth, normWave, normLande):
        """
        Returns the calculated LSD weights of all the lines in the mask
        (includes lines with both the iuse flag 0 and 1).
        
        This assumes the Stokes I lines are weighted as depth,
        and Stokes V is weighted as depth*wavelength*Lande factor.
        
        :param normDepth: the normalizing line depth for the mask/LSD profile
        :param normWave: the normalizing wavelength (nm) for
                         the mask/LSD profile
        :param normLande: the normalizing effective Lande factor for
                          the mask/LSD profile
        :return: weightI, weightV (as arrays)
        """
        weightI = self.depth / normDepth
        weightV = self.depth*self.wl*self.lande / (normDepth*normWave*normLande)
        return weightI, weightV
    
    def save(self, fname):
        """
        Save the line Mask to a text file, in Donati's and LSDpy format.
        
        :param fname: the file name to output the Mask to
        """
        
        nlines = self.wl.shape[0]
        with open(fname, 'w') as oFile:
            oFile.write('{:d}\n'.format(nlines))
            for i in range(nlines):
                oFile.write('{:9.4f} {:6.2f} {:6.3f} {:6.3f} {:6.3f} {:2d}\n'.format(
                    self.wl[i], self.element[i], self.depth[i],
                    self.excite[i], self.lande[i], self.iuse[i]))

        return
    
    def clean(self, regions):
        '''
        Remove lines inside a set of exclude regions from a mask.
        
        Takes an ExcludeMaskRegions object, and returns a Mask object with
        the iuse values set to zero for spectral lines with wavelengths
        inside these regions. 
        
        :param regions: An ExcludeMaskRegions object
        :rtype: Mask
        '''
        # Making a copy of the mask (or use copy.deepcopy(self) )
        mask_clean = Mask(self.wl.copy(), self.element.copy(),
                          self.depth.copy(), self.excite.copy(),
                          self.lande.copy(), self.iuse.copy())
        nregions = len(regions)
        for i in range(0,nregions):
            is_in = np.logical_and( (self.wl>=regions[i].start),
                                    (self.wl<=regions[i].stop) )
            mask_clean.iuse[is_in] = 0

        return mask_clean

def read_mask(fname):
    """
    Read in an LSD line mask file and return a Mask object.

    The mask file will have one line of header and columns of:  

    * Wavelength (nm)
    * Atomic number + (ionization state)*0.01
    * Line depth
    * Excitation potential of the lower level (eV)
    * Effective Lande factor
    * Flag for whether the line is used (1=use, 0=skip).

    :param fname: the name of the file to read.
    :rtype: Mask
    """
    tmpMask = np.loadtxt(fname, skiprows=1, unpack=True)
    
    #Sort the line mask so wavelength is always increasing
    ind = np.argsort(tmpMask[0,:], kind='stable')
    
    wl = tmpMask[0, ind]
    element = tmpMask[1, ind]
    depth = tmpMask[2, ind]
    excite = tmpMask[3, ind]
    lande = tmpMask[4, ind]
    iuse = tmpMask[5, ind].astype(int)
    
    return Mask(wl, element, depth, excite, lande, iuse)

###################################
###################################

class ExcludeMaskRegions:
    """
    Class for a region object that records spectral regions to exclude from a Mask.

    Usually contrains arrays of:  
    
    * start - starting wavelengths for the regions
    * stop - ending wavelengths for the regions
    * type - optionally, text comments for the type of region \
             (inside a numpy array with dtype=object)
    """

    def __init__(self, start, stop, type):
        self.start = start
        self.stop = stop
        self.type = type

    def __getitem__(self, key):
        """
        Returns a region object with only the values at the specified index(s).

        :param key: the index or slice being checked
        :rtype: ExcludeMaskRegions
        """
        return ExcludeMaskRegions(self.start[key], self.stop[key],
                                  self.type[key])

    def __setitem__(self, key, newval):
        """
        Sets all values of the Mask at the specified location
        equal to the input mask's values.

        :param key: the index or slice being overwritten
        :param newval: Mask whose values are to replace the overwritten ones
        """ 
        if not(isinstance(newval, ExcludeMaskRegions)):
            raise TypeError()
        else:
            self.start[key] = newval.start
            self.stop[key] = newval.stop
            self.type[key] = newval.type

    def __len__(self):
        """
        Returns the number of lines in the mask
        """
        return len(self.start)

    def __add__(self, other):
        """
        Concatenate two ExcludeMaskRegions objects.
        
        :param other: an ExcludeMaskRegions object to combine with this one
        :rtype: ExcludeMaskRegions
        """
        start = np.concatenate([self.start, other.start])
        stop = np.concatenate([self.stop, other.stop])
        type = np.concatenate([self.type, other.type])
        return ExcludeMaskRegions(start, stop, type)

    def save(self, fname):
        """
        Save the ExcludeMaskRegions object to a text file.

        The file contains one line for each region, with the start wavelength,
        end wavelength, and a type comment.
        
        :param fname: the file path/name
        """
        with open(fname, 'w') as ofile:
            for item in self:
                ofile.write('{:.4f} {:.4f} {}\n'.format(item.start, item.stop,
                                                        item.type))
        return

    def to_dict(self):
        """
        Return the ExcludeMaskRegions as a dictionary. 
        This is useful to transform ExcludeMaskRegions objects
        to Panda dataframes.
        """
        return {'start':self.start, 'stop':self.stop,'type':self.type}

def read_exclude_mask_regions(fname):
    """
    Read in an ExcludeMaskRegions file into an ExcludeMaskRegions object. 

    The file should be a text file with one line for each region,
    containing the start wavelength, end wavelength, and the type comment.
    
    :param fname: the path/name of the file
    :rtype: ExcludeMaskRegions
    """
    start = []
    stop = []
    type = []
    with open(fname, 'r') as f:
        nLines = 0
        for txtLine in f:
            split = txtLine.split() # split by WS
            start.append(split[0]) 
            stop.append(split[1])
            # join all the rest with ' ' in case there was multiple words
            type.append(' '.join(split[2:]))
    
    start = np.asarray(start, dtype=float)
    stop = np.asarray(stop, dtype=float)
    
    return ExcludeMaskRegions(start, stop, np.array(type, dtype=object))

def get_Balmer_regions_default(velrange=500):
    '''
    Generate an ExcludeMaskRegions object with regions around Balmer H-lines
    (alpha to epsilon) up to a given radial velocity,
    and a region that excludes the Balmer jump (from 360-392 nm).
    
    :param velrange: velocity range around the H-line to be excluded
                     (default 500 km/s)
    :rtype: ExcludeMaskRegions
    '''

    c = 299792.458 # Speed of light in km/s
    # Mask should be in nm
    wavelengths = [
        656.281,
        486.14,
        434.05,
        410.17,
        397.01
    ]
    types = [
            'Halpha',
            'Hbeta',
            'Hgamma',
            'Hdelta',
            'Hepsilon',
            'Hjump'
    ]
    start = []
    stop = []
    for w in wavelengths:
        start.append(-1*velrange/c*w + w)
        stop.append(velrange/c*w + w)

    # Adding the Balmer jump
    start.append(360.0)
    stop.append(392.0)

    return ExcludeMaskRegions(np.array(start), np.array(stop),
                              np.array(types,dtype=object))

def get_telluric_regions_default():
    '''
    Generate an ExcludeMaskRegions object with regions containing
    heavy telluric line contamination in the optical.
    
    :rtype: ExcludeMaskRegions
    '''
    start = np.array([587.5,627.5,684.0,717.0,757.0,790.0,809.0])  # nm
    stop   = np.array([592.0,632.5,705.3,735.0,771.0,795.0,990.0])  # nm

    return ExcludeMaskRegions(start, stop,
                              np.array(['telluric']*len(start),dtype=object))


###################################
###################################

def make_mask(lineListFile, outMaskName=None, depthCutoff=0.0, 
              wlStart=None, wlEnd=None, landeStart=None, landeEnd=None,
              elementsUsed=[], elementsExclude=[],
              atomsOnly=True, includeNoLande=False, defaultLande=1.0):
    """
    Generate an LSD line mask from a VALD line list and save it.
    This is the main interface for generating new line masks with Python scrips.
    
    This generates an LSD line mask from atomic lines in the list,
    and estimates Lande factors for lines where possible.  Lines for which 
    no Lande factor can be estimated are omitted by default.
    This assumes input line list wavelengths in A, and returns a mask in nm.
    
    :param lineListFile: The name of the file containing the line list
                         (in the VALD3 'extract stellar' 'long' format)
    :param outMaskName: The name of the to write the mask to
                        (set this to None to avoid saving a file, default)
    :param depthCutoff: Only include lines in the mask deeper than this value
                        (defaults to 0, all lines included)
    :param wlStart: Optionally, only use lines with wavelengths above this
                    (note: value in nm!)
    :param wlEnd: Optionally, only use lines with wavelengths below this
                  (note: value in nm!)
    :param landeStart: Optionally, only use lines with effective Lande factors
                       above this
    :param landeEnd: Optionally, only use lines with effective Lande factors
                     below this
    :param elementsUsed: Optionally, provide a list of elements to include
                         in the line mask. This should be a list of strings
                         with the element symbols, e.g. ['C', 'O', 'Si', 'Fe']
                         (an empty list or a list starting with 'all' will
                         include all elements)
    :param elementsExclude: Optionally, provide a list of elements to exclude
                            from the line mask. Also a list of strings with
                            the element symbols.
    :param atomsOnly: Only include atomic lines (no molecular lines) and
                      exclude H lines (defaults to True)
    :param includeNoLande: Include lines in the mask even if no Lande factor
                           can be estimated for them (defaults to False)
    :param defaultLande: The value of the effective Lande factor to use if
                         no value can be estimated (only used if
                         includeNoLande = True)
    :rtype: Mask
    """

    from . import lineList as lineListLib
    lineList = lineListLib.read_VALD(lineListFile)
    
    mask = convert_list_to_mask(lineList, depthCutoff=depthCutoff,
                                atomsOnly=atomsOnly,
                                includeNoLande=includeNoLande,
                                defaultLande=defaultLande)
    
    mask = filter_mask(mask, depthCutoff=depthCutoff,
                       wlStart=wlStart, wlEnd=wlEnd,
                       landeStart=landeStart, landeEnd=landeEnd,
                       elementsUsed=elementsUsed,
                       elementsExclude=elementsExclude)

    if outMaskName is not None:
        mask.save(outMaskName)
    return mask

    
def convert_list_to_mask(lineList, depthCutoff=0.0, atomsOnly = True,
                         includeNoLande = False, defaultLande=1.0):
    """
    Convert a lineList to an LSD line mask.
    
    This estimates Lande factors when VALD doesn't have a known value.
    This can also automatically reject any H lines and any molecular lines.
    Note, this assumes an input line list in A, and returns a mask in nm.
    See also the make_mask() function.
    
    :param lineList: The line list object to be converted to a mask
    :param depthCutoff: Only include lines deeper than this depth cutoff
                        (defaults to 0, all lines included)
    :param atomsOnly: Flag to use only atomic lines (no molecular lines)
                      and exclude H lines (defaults to True)
    :param includeNoLande: Flag to include lines in the mask even if no Lande
                           factor can be estimated for them (defaults to False)
    :param defaultLande: The value of the effective Lande factor to use for
                         lines where no value can be estimated
                         (only used if includeNoLande = True)
    :rtype: Mask
    """
    from . import lineList as lineListLib
    elements=('H' ,'He','Li','Be','B' ,'C' ,'N' ,'O' ,'F' ,'Ne','Na','Mg',
              'Al','Si','P' ,'S' ,'Cl','Ar','K' ,'Ca','Sc','Ti','V' ,'Cr',
              'Mn','Fe','Co','Ni','Cu','Zn','Ga','Ge','As','Se','Br','Kr',
              'Rb','Sr','Y' ,'Zr','Nb','Mo','Tc','Ru','Rh','Pd','Ag','Cd',
              'In','Sn','Sb','Te','I' ,'Xe','Cs','Ba','La','Ce','Pr','Nd',
              'Pm','Sm','Eu','Gd','Tb','Dy','Ho','Er','Tm','Yb','Lu','Hf',
              'Ta','W' ,'Re','Os','Ir','Pt','Au','Hg','Tl','Pb','Bi','Po',
              'At','Rn','Fr','Ra','Ac','Th','Pa','U' )
    nLines = lineList.nLines
    mask = Mask(np.zeros(nLines), np.zeros(nLines), np.zeros(nLines),
                np.zeros(nLines), np.zeros(nLines),
                np.zeros(nLines,dtype=int))

    skippedIon = []
    missingLandeIon = []
    missingLandeNum = 0
    j = 0
    #Loop through the line list making sure to use good data
    for i in range(lineList.nLines):
        element = lineList.ion[i].split()[0]
        ion = lineList.ion[i].split()[1]

        if (lineList.depth[i] >= depthCutoff) and (
            (atomsOnly == False) or (element in elements and element != 'H')):
            #Estimate an effective Lande factor if necessary, VALD flags 
            #levels without a known Lande factor using the value 99.0
            landeEff = lineList.landeEff[i]
            if lineList.landeEff[i] >= 90.:
                landeLo = lineList.landeLo[i]
                if landeLo >= 90.:
                    landeLo = lineListLib.estimateLande(
                        lineList.Jlo[i], lineList.configLo[i])
                landeUp = lineList.landeUp[i]
                if landeUp >= 90.:
                    landeUp = lineListLib.estimateLande(
                        lineList.Jup[i], lineList.configUp[i])
                #If we have managed to estimate Lande factors
                if landeLo < 90. and landeUp < 90.:
                    landeEff = lineListLib.getEffectiveLande(
                        landeLo, landeUp, lineList.Jlo[i], lineList.Jup[i])

            #If we can't calculate a Lande factor but still want to use the line
            if landeEff >= 90. and includeNoLande == True:
                landeEff = defaultLande

            #For lines with a Lande factor
            if landeEff < 90.:
                #Save the data for good lines
                mask.wl[j]     = lineList.wl[i]
                mask.depth[j]  = lineList.depth[i]
                mask.excite[j] = lineList.Elo[i]
                mask.lande[j]  = landeEff
                mask.iuse[j]   = 1
                #The element code here is the atomic number plus 
                #the ionization state/100, so Fe II becomes 26.01
                try:
                    mask.element[j] = float(elements.index(element)+1) \
                        + (float(ion)-1)*0.01
                except ValueError:
                    mask.element[j] = 0.00
                
                j += 1
            else:
                #Track lines not used due to missing Lande factors
                missingLandeNum += 1
                if not lineList.ion[i] in missingLandeIon:
                    missingLandeIon += [lineList.ion[i]]
        elif (lineList.depth[i] >= depthCutoff):
            #Track lines not used due unsupported species
            if not lineList.ion[i] in skippedIon:
                skippedIon += [lineList.ion[i]]

    if missingLandeNum > 0:
        print('missing Lande factors for {:} lines (skipped) from:'.format(
            missingLandeNum))
        print(missingLandeIon)
    if len(skippedIon) > 0:
        print('skipped all lines for species:')
        print(skippedIon)
    
    #Convert wavelengths from A to values in nm
    mask.wl = mask.wl*0.1
    #Remove blank entries in the line mask
    mask = mask.prune()
    
    return mask


def filter_mask(mask, depthCutoff = 0.0, wlStart = None, wlEnd = None,
               landeStart = None, landeEnd = None,
               elementsUsed = [], elementsExclude = []):
    """
    Remove lines from an LSD line mask, based on their depth, wavelength,
    effective Lande factor, or element.

    Usually, only one of elementsUsed and elementsExclude should be used.
    Both can be used, in that case only elements in elementsUsed are included,
    and then any elements in elementsExclude are removed from that.
    (Filtering by element currently only works for atomic lines,
    not molecular lines)

    :param mask: The line mask object to filter lines out of
    :param depthCutoff: Only include lines in the mask deeper than this value
                        (defaults to 0, all lines included)
    :param wlStart: Optionally, only use lines with wavelengths above this
                    (note: value in nm!)
    :param wlEnd: Optionally, only use lines with wavelengths below this
                  (note: value in nm!)
    :param landeStart: Optionally, only use lines with effective Lande factors
                       above this
    :param landeEnd: Optionally, only use lines with effective Lande factors
                     below this
    :param elementsUsed: Optionally, provide a list of elements to include
                         in the line mask. This should be a list of strings
                         with the element symbols, e.g. ['C', 'O', 'Si', 'Fe']
                         (an empty list or a list starting with 'all' will
                         include all elements)
    :param elementsExclude: Optionally, provide a list of elements to exclude
                            from the line mask. Also a list of strings with
                            the element symbols.
    :rtype: Mask
    """
    
    elements=('H' ,'He','Li','Be','B' ,'C' ,'N' ,'O' ,'F' ,'Ne','Na','Mg',
              'Al','Si','P' ,'S' ,'Cl','Ar','K' ,'Ca','Sc','Ti','V' ,'Cr',
              'Mn','Fe','Co','Ni','Cu','Zn','Ga','Ge','As','Se','Br','Kr',
              'Rb','Sr','Y' ,'Zr','Nb','Mo','Tc','Ru','Rh','Pd','Ag','Cd',
              'In','Sn','Sb','Te','I' ,'Xe','Cs','Ba','La','Ce','Pr','Nd',
              'Pm','Sm','Eu','Gd','Tb','Dy','Ho','Er','Tm','Yb','Lu','Hf',
              'Ta','W' ,'Re','Os','Ir','Pt','Au','Hg','Tl','Pb','Bi','Po',
              'At','Rn','Fr','Ra','Ac','Th','Pa','U' )

    if wlStart is None: wlStart = 0.0
    if wlEnd is None: wlEnd = 1e10
    if landeStart is None: landeStart = -1e10
    if landeEnd is None: landeEnd = 1e10,
    
    #Remove lines that are outside the requested range in wavelength, Lande factor, or depth.
    indRemove = ((mask.depth < depthCutoff)
                 | (mask.wl < wlStart) | (mask.wl > wlEnd) 
                 | (mask.lande < landeStart) | (mask.lande > landeEnd))
    
    #If there is also a list of specific elements to use
    if type(elementsUsed) != type([]): print('ERROR: list of elements to include not of type list')
    if elementsUsed != [] and elementsUsed[0].lower() != 'all':
        indKeep = np.zeros_like(mask.iuse, dtype=bool)
        #Set a flag to only keep elements in the requested list
        for elUse in elementsUsed:
            try: #get the atomic number for this element
                elNumber = elements.index(elUse.strip())+1
            except ValueError:
                raise ValueError('Unrecognized element {:} in elementsUsed'.format(elUse))
            indKeep = indKeep | (mask.element.astype(int) == elNumber)
        indRemove = indRemove | np.logical_not(indKeep)
    
    #Or if there is also a list of specific elements to exclude from the mask
    if type(elementsExclude) != type([]): print('ERROR: list of elements to exclude not of type list')
    if elementsExclude != [] :
        #if elementsUsed != [] and elementsUsed[0].lower() != 'all':
        #    print('Warning: got both a list of elements to include and to exclude. ')
        indExclude = np.zeros_like(mask.iuse, dtype=bool)
        #Set a flag to remove elements in the exclude list
        for elCut in elementsExclude:
            try: #get the atomic number for this element
                elNumber = elements.index(elCut.strip())+1
            except ValueError:
                raise ValueError('Unrecognized element {:} in elementsExclude'.format(elUse))
            indExclude = indExclude | (mask.element.astype(int) == elNumber)
        indRemove = indRemove | indExclude
    
    #Apply the filtering
    mask.iuse[indRemove] = 0
    
    #Remove blank entries in the line mask
    mask = mask.prune()
    
    return mask



#!/usr/bin/python3
#
# Generate an LSD line mask from a VALD line list.

import argparse
try:
    import specpolFlow as pol
except ModuleNotFoundError:
    #If specpolFlow is not installed or in the python path
    #try guessing the location and adding it temporarily
    import sys, os
    loc0 = os.path.dirname(__file__) 
    sys.path.insert(0, os.path.join(loc0, '..', '..'))
    import specpolFlow as pol

def make_mask_cli():
    """
    Main function for running make_mask as a terminal program.

    Takes no arguments, but uses command line parameters instead.
    """
    #Take input and output file names as command line arguments,
    #with some additional optional control parameters.
    parser = argparse.ArgumentParser(description='Generate an LSD line mask from a VALD line list. When Lande factors are not in the line list they will be estimated if possible.')
    parser.add_argument("lineList", help="Line list from VALD. This file should be an 'extract stellar' line list in the 'long' format.")
    parser.add_argument("mask", nargs='?', default='mask.dat', help='Save the line mask to this file.')
    parser.add_argument("-d", "--depth", default='0', help='Optional number, only include lines deeper than this depth cutoff in the mask (defaults to 0, all lines included)')
    parser.add_argument("-w1", default='0', help='Optional number, starting wavelength for lines included in the mask (in nm)')
    parser.add_argument("-w2", default='1e10', help='Optional number, ending wavelength for lines included in the mask (in nm)')
    parser.add_argument("-g1", default='-1e10', help='Optional number, lower cutoff in effective Lande factor for lines included in the mask')
    parser.add_argument("-g2", default='1e10', help='Optional number, upper cutoff in effective Lande factor')
    parser.add_argument("-e", "--elements", default='', help="Optional, elements to include in the mask, as list symbols. For multiple elements enclose in quotes, e.g. 'C O Fe Ni'. By default all elements are used.")
    parser.add_argument("-ex", "--excludeEl", default='', help="Optional, elements to exclude from the mask, as a list symbols. For multiple elements enclose in quotes, e.g. 'C O Fe Ni'. By default all elements are used.")
    args = parser.parse_args()
    
    llistName = args.lineList
    maskName = args.mask
    depthCutoff = float(args.depth)
    wl1 = float(args.w1)
    wl2 = float(args.w2)
    lande1 = float(args.g1)
    lande2 = float(args.g2)
    elementsUsed = args.elements.split()
    elementsExclude = args.excludeEl.split()
    
    #Run the line mask generating code
    print('reading line list from', llistName)
    mask = pol.make_mask(llistName, outMaskName = maskName, 
                         depthCutoff = depthCutoff, wlStart = wl1, wlEnd = wl2,
                         landeStart = lande1, landeEnd = lande2,
                         elementsUsed = elementsUsed,
                         elementsExclude = elementsExclude,
                         includeNoLande = False)
    print('mask saved to', maskName)
    return

#For running this script as a terminal program
if __name__ == "__main__":
    make_mask_cli()



"""
Tools for manipulating spectra, typically spectropolarimetric observations.
"""

import numpy as np
import copy

###################################

class Spectrum:
    """
    Contains an observed spectrum, usually spectropolarimetric data.

    Usually contains arrays: 
     
    * wl - wavelengths
    * specI - Stokes I spectrum
    * specV - polarized spectrum, usually Stokes V
    * specN1 - the first polarimetric null spectrum
    * specN2 - the second polarimetric null spectrum
    * specSig - the formal uncertainties, which apply to the other spectra
    """

    def __init__(self, wl, specI, specV, specN1, specN2, specSig, header=None):
        
        self.header=header
        self.wl = wl
        self.specI=specI
        self.specV=specV
        self.specN1=specN1
        self.specN2=specN2
        self.specSig=specSig
        
    def __getitem__(self, key):
        """
        Returns a Spectrum object with only the values at the specified index(s)

        :param key: the index or slice being checked
        :rtype: Spectrum
        """
        wl_s = self.wl[key]
        specI_s = self.specI[key]
        specSig_s = self.specSig[key]
        specV_s = self.specV[key]
        specN1_s = self.specN1[key]
        specN2_s = self.specN2[key]
        #The header may be None but that is ok.
        slice_spec = Spectrum(wl_s, specI_s, specV_s, specN1_s, specN2_s,
                              specSig_s, header=self.header)
        return slice_spec

    def __setitem__(self, key, newval):
        """
        Sets all values of the Spectrum at the specified location equal
        to the input Spectrum's values.

        :param key: the index or slice being overwritten
        :param newval: Spectrum whose values are to replace the overwritten ones
        """
        if not(isinstance(newval, Spectrum)):
            raise TypeError()
        else:
            self.wl[key] = newval.wl
            self.specI[key] = newval.specI
            self.specSig[key] = newval.specSig
            self.specV[key] = newval.specV
            self.specN1[key] = newval.specN1
            self.specN2[key] = newval.specN2

    def __len__(self):
        return len(self.wl)

    def concatenate(self, *args):
        """
        Combine this Spectrum with other Spectrum objects, passed as arguments,
        concatenating them in order, into a new Spectrum object.

        :param args: other Spectrum objects (or a list or tuple of them)
                     to concatenate with this one
        :rtype: Spectrum
        """
        cat_header = self.header
        cat_wl = self.wl
        cat_specI = self.specI
        cat_specV = self.specV
        cat_specN1 = self.specN1
        cat_specN2 = self.specN2
        cat_specSig = self.specSig
        args2 = args
        #Check if the user passed a list or tuple of spectra (unwrap it)
        if len(args) > 0:
            if isinstance(args[0], list) or isinstance(args[0], tuple):
                if isinstance(args[0][0], Spectrum):
                    args2 = args[0]
        for arg in args2:
            if not isinstance(arg, Spectrum):
                raise ValueError('concatenate can only use Spectrum objects!')
            cat_wl = np.concatenate((cat_wl, arg.wl))
            cat_specI = np.concatenate((cat_specI, arg.specI))
            cat_specV = np.concatenate((cat_specV, arg.specV))
            cat_specN1 = np.concatenate((cat_specN1, arg.specN1))
            cat_specN2 = np.concatenate((cat_specN2, arg.specN2))
            cat_specSig = np.concatenate((cat_specSig, arg.specSig))
        cat_spec = Spectrum(cat_wl, cat_specI, cat_specV, cat_specN1,
                            cat_specN2, cat_specSig, header=cat_header)
        return cat_spec

    def coadd(self, *args):
        """
        coadd this spectrum with other spectra

        Uses the the wavelength grid of this spectrum for the coadded spectrum.
        Other spectra are interpolated onto this spectrum's wavelengths before
        coadding.  Coadding here essentially averages spectra weighted
        by 1/sigma**2. This assumes the spectra are continuum normalized,
        and have reliable uncertainties.

        Warning: regions with order overlap, or where wavelength goes backwards,
        will produce incorrect results with this routine.  We strongly
        recommend using the merge_orders function before using this routine.
        
        :param args: other Spectrum objects (or a list or tuple of them)
                     to coadd with this one
        :rtype: Spectrum
        """
        #Set up the weighted average using self as the first entry
        #weight by 1/sigma**2, and save the sum of the weights
        spec = copy.deepcopy(self) #work on a copy (not self!)
        weight = 1./self.specSig**2
        spec.specI = self.specI*weight
        spec.specV = self.specV*weight
        spec.specN1 = self.specN1*weight
        spec.specN2 = self.specN2*weight
        totalWeight = weight.copy()

        args2 = args
        #Check if the user passed a list or tuple of spectra (unwrap it)
        if len(args) > 0:
            if isinstance(args[0], list) or isinstance(args[0], tuple):
                if isinstance(args[0][0], Spectrum):
                    args2 = args[0]
        for arg in args2:
            if not isinstance(arg, Spectrum):
                raise ValueError('coadd can only use Spectrum objects!')
            if np.any(arg.specSig <= 0.0):
                raise ValueError("coadd requires uncertainties "
                                 "(>0) for all points!")
            if not np.all(np.diff(arg.wl) > 0):
                print('Warning: in Spectrum.coadd, coadding spectra with order'
                      ' overlap or decreasing wavelength.  This will cause'
                      ' incorrect results in those regions.')
            #Interpolate this spectrum onto the reference wavelengths
            bspecI   = np.interp(spec.wl, arg.wl, arg.specI)
            bspecV   = np.interp(spec.wl, arg.wl, arg.specV)
            bspecN1  = np.interp(spec.wl, arg.wl, arg.specN1)
            bspecN2  = np.interp(spec.wl, arg.wl, arg.specN2)
            bspecSig = np.interp(spec.wl, arg.wl, arg.specSig)

            #Add this spectrum to the weighted sums
            weight = 1./bspecSig**2
            spec.specI += bspecI*weight
            spec.specV += bspecV*weight
            spec.specN1 += bspecN1*weight
            spec.specN2 += bspecN2*weight
            totalWeight += weight

        #Once all spectra have been added, complete the weighted average
        spec.specI /= totalWeight
        spec.specV /= totalWeight
        spec.specN1 /= totalWeight
        spec.specN2 /= totalWeight
        spec.specSig = np.sqrt(1/totalWeight)
        return spec
            
    def individual_line(self, lambda0, lwidth):
        '''
        Select an individual line in the spectrum and return it
        as an LSD profile object

        :param lambda0: wavelength of the line (same units as self.wl)
        :param lwidth: distance from the line center, in wavelength,
                       for the wavelength window used for the line profile.
                       One element: same distance on each side of line center.
                       Two elements: distance to the left and right of
                       line center.
        :rtype: LSD
        '''
        #This is nearly a circular import, since profileLSD imports obsSpec
        #Maybe move this to a stand alone function in profileLSD?
        from .profileLSD import LSD
        
        # Select observed line
        if isinstance(lwidth, list) or isinstance(lwidth, tuple):
            if len(lwidth) == 1:
                p_lwidth = [lambda0 - lwidth[0], lambda0 + lwidth[0]]
            elif len(lwidth) == 2:
                p_lwidth = [lambda0 - lwidth[0], lambda0 + lwidth[1]]
            else:
                print('lwidth has too many elements (need one or two)')
                raise ValueError('lwidth has too many elements: '
                                 '{:} (need 1 or 2)'.format(len(lwidth)))
        else:
            p_lwidth = [lambda0 - lwidth, lambda0 + lwidth]
        
        obs_line = self[(self.wl >= p_lwidth[0]) & (self.wl <= p_lwidth[1])]

        # Now we convert wavelengths to velocity space
        c = 299792.458  #speed of light in km/s
        vel = c*(obs_line.wl-lambda0)/lambda0

        prof = LSD(vel, obs_line.specI, obs_line.specSig, obs_line.specV, 
                   obs_line.specSig, obs_line.specN1, obs_line.specSig,
                   header=obs_line.header)
        return prof 

    def save(self, fname, saveHeader=True):
        '''
        Write the Spectrum into a text .s file in a LibreESPRIT style format.
        Optionally skip writing the two lines of header.

        :param saveHeader: optional flag, skip writing the header if False
        '''

        #Note, the LibreESPRIT .s format header counts the number of columns
        #data columns not counting the first wavelength column
        ncols = 5
        #Support 3 column (intensity only) spectra
        if np.all(self.specV == 0.):
            ncols = 2
            #and 2 column (intensity only, no errorbars) spectra
            if np.all(self.specSig == 0):
                ncols = 1
        
        with open(fname, 'w') as f:
            #Optionally write 2 lines of header            
            if saveHeader:
                if self.header is None:
                    f.write('*** Spectrum of\n')
                else:
                    f.write(self.header)
                    #Make sure there is a line break after the first header text
                    if self.header[-1] != '\n': f.write('\n')
                f.write('{:7d} {:1d}\n'.format(int(self.wl.size), ncols))
            
            if ncols == 5:
                for i in range(self.wl.size):
                    f.write('{:10.4f} {:11.4e} {:11.4e} {:11.4e} {:11.4e} {:11.4e}\n'.format(
                        self.wl[i], self.specI[i], self.specV[i],
                        self.specN1[i], self.specN2[i], self.specSig[i]))
            elif ncols == 2:
                for i in range(self.wl.size):
                    f.write('{:10.4f} {:11.4e} {:11.4e}\n'.format(
                        self.wl[i], self.specI[i], self.specSig[i]))
            elif ncols == 1:
                for i in range(self.wl.size):
                    f.write('{:10.4f} {:11.4e}\n'.format(
                        self.wl[i], self.specI[i]))
        return

    def get_orders(self, ignoreGaps=False):
        """
        Split an observed echelle spectrum into individual echelle orders.
        
        This splits a concatenated spectrum into echelle orders, based on overlap
        (places where wavelength decreases), and optionally gaps in wavelength
        between orders.

        :param ignoreGaps: if True then regions separated by gaps in wavelength
                           are treated as one order, otherwise gaps are used as
                           order edges
        :return: a list of Spectrum objects, one for each identified order
        """
        #Define spectral order edges by a step backwards in wavelength (velocity)
        #or a step forward in velocity more than 10x the average velocity step size.
        #(use velocity rather than wavelength since pixel size in velocity is
        # more consistent across a spectrum)
        c = 299792.458  #speed of light in km/s
        gapSize = 20.
        velSteps = (self.wl[1:-1] - self.wl[0:-2])/self.wl[1:-1]*c
        meanVelStep = np.mean(velSteps)
        if ignoreGaps:
            orderEdges = velSteps < 0.
        else:
            orderEdges = np.logical_or(velSteps < 0., velSteps > gapSize*meanVelStep)
        indOrderEdges =  np.nonzero(orderEdges)[0] #last point in a spectral order
        numOrders = indOrderEdges.shape[0] + 1

        #Slice the orders out of this spectrum and put them in a new list
        scp = copy.deepcopy(self) #start with a copy (don't just modify self!)
        orders = []
        indLast = 0
        for i in range(numOrders - 1):
            orders.append(scp[indLast : indOrderEdges[i]+1])
            indLast = indOrderEdges[i]+1
        orders.append(scp[indLast :])
        
        return orders

    def get_orders_in_range(self, wl1, wl2=None, ignoreGaps=False):
        """
        Split an observed echelle spectrum into individual orders,
        and return orders that include the specified wavelength range.

        This splits a concatenated spectrum into echelle orders, based on overlap
        (places where wavelength decreases), and optionally gaps in wavelength
        between orders.  Returning the order, or orders, that include any part
        of the specified wavelength range (wl1 and wl2 can be identical).

        :param wl1: the start of the wavelength range to include
        :param wl2: the end of the wavelength range to include
                    (if not provided, orders including just the point in
                    wavelength wl1 will be returned)
        :param ignoreGaps: if True then regions separated by gaps in wavelength
                           are treated as one order, otherwise gaps are used as
                           order edges
        :return: a list of Spectrum objects, one for each relevant order
        """
        if wl2 is None:  #If we just want to check one point in wavelength
            wl2 = wl1
        if wl1 > wl2:
            raise ValueError('Start wavelength must be smaller than end wavelength'
                             ' got {:} {:}'.format(wl1, wl2))
        
        allOrders = self.get_orders(ignoreGaps)
        retOrders = []
        for order in allOrders:
            #if the order starts before the desired range ends,
            #and the order ends after the desired range starts,
            #then some (or all) of it should be in range
            if(wl2 > order.wl[0] and wl1 < order.wl[-1]): 
                retOrders.append(order)

        return retOrders

    def merge_orders(self, mode='trim', midpoint=0.5):
        """
        Simple merging of spectral orders, for echelle spectra.

        The default mode 'trim' simply uses one order up to the midpoint
        of an overlap region, and then uses the next order past the midpoint.
        This is more robust against continuum normalization errors but
        doesn't optimize the total signal-to-noise.
        
        The mode 'coadd' attempts to coadd orders in the overlap region.
        It interpolates the second order onto the wavelength grid of the
        first order, then averages the spectra weighted by 1/sigma**2.
        This optimizes the total signal-to-noise but is vulnerable 
        to continuum normalization errors at the edges of orders.
        This mode requires reliable uncertainties.

        :param mode: choice of 'trim', 'coadd'
        :param midpoint: for mode 'trim', the fraction of the way through
                         an overlap region treated as the midpoint
        :rtype: Spectrum
        """

        orders = self.get_orders(ignoreGaps=True)
        numOrders = len(orders)

        if mode.lower() == 'trim':
            #Merge by splitting orders at the midpoint of their overlap
            wlStartMid = 0.0
            for i in range(numOrders):
                #find the midpoint of the overlap between this and the next order
                if i == numOrders - 1: #for the last order include to the end
                    wlEndMid = orders[i].wl[-1]
                else:
                    wlEndMid = (orders[i].wl[-1]*midpoint
                                + orders[i+1].wl[0]*(1. - midpoint))
                #use this order between this midpoint and the previous midpoint
                induse = (orders[i].wl > wlStartMid) & (orders[i].wl <= wlEndMid)

                #concatenate the used parts of the orders
                if i == 0: #(or just use this order if it's the 1st)
                    specM = orders[i][induse]
                else:
                    specM = specM.concatenate(orders[i][induse])
                wlStartMid = wlEndMid
            
        elif mode.lower() == 'coadd':
            #Merge by coadding orders inside the overlap region
            if np.any(self.specSig <= 0.0):
                raise ValueError("merge_orders in mode 'coadd' requires uncertainties "
                                 "(>0) for all points!")
            for i in range(numOrders):
                #use this order's wavelength grid in the overlap with the next order
                #(assuming all wavelength solutions are equally good)
                if i < numOrders - 1:
                    #identify the overlap range between orders
                    ord1 = orders[i]
                    ord2 = orders[i+1]
                    indO1 = ord1.wl >= ord2.wl[0]
                    indO2 = ord2.wl <= ord1.wl[-1]
                    #Include one extra point from the next order for interpolation
                    indO2[np.nonzero(np.logical_not(indO2))[0][0]] = True
                    
                    #coadd the region in the overlap
                    ord1[indO1] = ord1[indO1].coadd(ord2[indO2])
                    
                #Save this order with coadded values,
                #except for the overlap with the previous order
                if i == 0: #(just use this order if it's the 1st)
                    specM = orders[i]
                else:
                    induse = orders[i].wl > orders[i-1].wl[-1]
                    specM = specM.concatenate(orders[i][induse])            
        else:
            raise ValueError("in merge_orders unrecognized mode '{:}'!".format(mode))
        return specM
    

def read_spectrum(fname, trimBadPix=False, sortByWavelength=False):
    """
    Read in the observed spectrum and save it.
    
    This follows the .s format from Donati's LibreESPRIT.
    Files can either have two lines of header or no header.
    This supports 6 column spectropolarimetric files
    (wavelength, I, V|Q|U, null1, null2, errors),
    and also 3 column spectra (wavelength, I, errors).

    :param fname: the name of the file to read.
    :param trimBadPix: optionally remove the more obviously bad pixels if True.
                       Removes pixels with negative flux or error bars of zero,
                       pixels with flux within 3sigma of zero (large errors),
                       and pixels with extremely large values.
    :param sortByWavelength: reorder the points in the spectrum to always
                             increase in wavelength, if set to True.
    :rtype: Spectrum
    """
    # Reading manually is often faster than np.loadtxt for a large files
    fObs = open(fname, 'r')
    #Check if the file starts with data or a header (assume 2 lines of header)
    line1 = fObs.readline()
    line2 = fObs.readline()
    line3 = fObs.readline()
    #assume at least the 3rd line contains real data
    ncolumns = len(line3.split())
    if ncolumns != 6 and ncolumns != 3 and ncolumns != 2:
        print('{:} column spectrum: unknown format!\n'.format(ncolumns))
        raise ValueError(('Reading {:} as an {:} column spectrum: '
                          'unknown format!').format(fname, ncolumns))
    
    if len(line1.split()) == ncolumns and len(line2.split()) == ncolumns:
        #If the column counts are consistent there may be no header
        try:
            #and if the first line starts with numbers, probably no header
            float(line1.split()[0])
            float(line1.split()[1])
            float(line2.split()[0])
            float(line2.split()[1])
            obs_header = None
            nHeader = 0
        except ValueError:
            #Otherwise assume there are two lines of header,
            #one line of with comment and a second with file dimensions,
            #but we figure that by reading the file.
            obs_header = line1
            nHeader = 2
    else:
        obs_header = line1
        nHeader = 2

    #Get the number of lines of data in the file
    nLines = 3 - nHeader #we've already read 3 lines
    for line in fObs:
        words = line.split()
        if len(words) == ncolumns:
            nLines += 1
        else:
            print('ERROR: reading observation, '
                  +'line {:}, {:} columns :\n{:}'.format(
                      nLines, len(words), line))

    obs = Spectrum(np.zeros(nLines), np.zeros(nLines), np.zeros(nLines),
                   np.zeros(nLines), np.zeros(nLines), np.zeros(nLines),
                   header=obs_header)
    
    #Rewind to start then advance the file pointer 2 lines
    fObs.seek(0)
    if obs_header is not None:
        fObs.readline()
        fObs.readline()
    #Then read the actual data of the file
    for i, line in enumerate(fObs):
        words = line.split()
        if (len(words) == ncolumns and ncolumns == 6):
            obs.wl[i] = float(words[0])
            obs.specI[i] = float(words[1])
            obs.specV[i] = float(words[2])
            obs.specN1[i] = float(words[3])
            obs.specN2[i] = float(words[4])
            obs.specSig[i] = float(words[5])
        elif (len(words) == ncolumns and ncolumns == 3):
            obs.wl[i] = float(words[0])
            obs.specI[i] = float(words[1])
            obs.specSig[i] = float(words[2])
        elif (len(words) == ncolumns and ncolumns == 2):
            obs.wl[i] = float(words[0])
            obs.specI[i] = float(words[1])
            
    fObs.close()

    #Optionally, remove bad pixels.
    if trimBadPix:
        use_flag = (obs.specI > 0.)
        if not np.all(obs.specSig == 0):  #(all zeros => no error column)
            use_flag = use_flag & (obs.specSig > 0.)
        use_flag = use_flag & (obs.specI > 3*obs.specSig)
        use_flag = use_flag & (obs.specI < 10*np.percentile(obs.specI, 99.9))
        obs = obs[use_flag]
    
    #Optionally, sort the observation so wavelength is always increasing
    if sortByWavelength:
        obs_ind = np.argsort(obs.wl)
        obs = obs[obs_ind]
    
    return obs

###################################



#!/usr/bin/python3
#
# Interactive tool for cleaning poor lines from an LSD line mask
# also can fit line depths in the mask.

import argparse

try:
    import specpolFlow as pol
except ModuleNotFoundError:
    #If specpolFlow is not installed or in the python path
    #try guessing the location and adding it temporarily
    import sys, os
    loc0 = os.path.dirname(__file__)
    sys.path.insert(0, os.path.join(loc0, '..', '..'))
    import specpolFlow as pol

def clean_mask_cli():
    """Main function for running cleanMaskUI as a terminal program.
    """
    parser = argparse.ArgumentParser(description='Interactively remove poor lines from an LSD line mask, and optionally modify line depths. This plots the mask, a comparison observation, and the model LSD spectrum (the convolution of the mask and LSD profile). For the model LSD spectrum calculation, some more useful parameters can be set within the graphical UI, and there is an option for displaying the resulting profile.')
    parser.add_argument('mask', help='The line mask to clean')
    parser.add_argument('observation', help='An observed spectrum for comparison')
    parser.add_argument('outName', nargs='?', default=None, help='Optional name for the output cleaned line mask, defaults to [input_file_name].clean')
    parser.add_argument('-e', '--exclude', default='excludeRanges.dat', help='Optional, a file of wavelength ranges to exclude from the line mask.  If the file does not exist default initial values will be used.')
    parser.add_argument('-b', '--batch', action='store_true', help='Run in batch mode, just apply the exclude regions from file, skip the interactive UI and LSD calculations.')
    args = parser.parse_args()

    maskName = args.mask
    obsName = args.observation
    outMaskName = args.outName
    excludeFileName = args.exclude
    batchMode = args.batch
    
    cleanMask, excludeRanges = pol.cleanMaskUI(maskName, obsName,
                                               outMaskName=outMaskName,
                                               inExcludeName=excludeFileName,
                                               outExcludeName=excludeFileName,
                                               batchMode=batchMode)
    return

#For running this Python script as a terminal program
if __name__ == "__main__":
    clean_mask_cli()
